"""Use Monte Carlo to propagate uncertainties"""

from multiprocessing import Pool

import comet_maths as cm
import numpy as np
import punpy.utilities.utilities as util

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


class LPUPropagation:
    def __init__(self, parallel_cores=0, Jx_diag=False, step=None, verbose=False):
        """
        Initialise Law of Propagation of Uncertainty Propagator

        :param parallel_cores: number of CPU to be used in parallel processing
        :type parallel_cores: int
        :param Jx_diag: Bool to indicate whether the Jacobian matrix can be described with semi-diagonal elements. With this we mean that the measurand has the same shape as each of the input quantities and the square jacobain between the measurand and each of the input quantities individually, only has diagonal elements. Defaults to False
        :rtype Jx_diag: bool, optional
        """

        self.parallel_cores = parallel_cores
        self.Jx_diag = Jx_diag
        self.step = step
        self.verbose = verbose

    def propagate_random(
        self,
        func,
        x,
        u_x,
        corr_x=None,
        param_fixed=None,
        corr_between=None,
        return_corr=False,
        return_Jacobian=False,
        repeat_dims=-99,
        corr_axis=-99,
        fixed_corr_var=False,
        output_vars=1,
        Jx=None,
        Jx_diag=None,
    ):
        """
        Propagate random uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Random uncertainties arise when there is no correlation between repeated measurements.
        It is possible (though rare) that there is a correlation in one of the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of random uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian matrix, defaults to False
        :type return_Jacobian: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param Jx: Jacobian matrix, evaluated at x. This allows to give a precomputed jacobian matrix, which could potentially be calculated using analytical prescription. Defaults to None, in which case Jx is calculated numerically as part of the propagation.
        :rtype Jx: array, optional
        :param Jx_diag: Bool to indicate whether the Jacobian matrix can be described with semi-diagonal elements. With this we mean that the measurand has the same shape as each of the input quantities and the square jacobain between the measurand and each of the input quantities individually, only has diagonal elements. Defaults to None, in which case the object value is used.
        :rtype Jx_diag: bool, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        if corr_x is None:
            corr_x = ["rand"] * len(x)
        for i in range(len(x)):
            if corr_x[i] is None:
                corr_x[i] = "rand"

        return self.propagate_standard(
            func,
            x,
            u_x,
            corr_x,
            param_fixed=param_fixed,
            corr_between=corr_between,
            return_corr=return_corr,
            return_Jacobian=return_Jacobian,
            repeat_dims=repeat_dims,
            corr_axis=corr_axis,
            fixed_corr_var=fixed_corr_var,
            output_vars=output_vars,
            Jx=Jx,
            Jx_diag=Jx_diag,
        )

    def propagate_systematic(
        self,
        func,
        x,
        u_x,
        corr_x=None,
        param_fixed=None,
        corr_between=None,
        return_corr=False,
        return_Jacobian=False,
        repeat_dims=-99,
        corr_axis=-99,
        fixed_corr_var=False,
        output_vars=1,
        Jx=None,
        Jx_diag=None,
    ):
        """
        Propagate systematic uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Systematic uncertainties arise when there is full correlation between repeated measurements.
        There is a often also a correlation between measurements along the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian matrix, defaults to False
        :type return_Jacobian: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param Jx: Jacobian matrix, evaluated at x. This allows to give a precomputed jacobian matrix, which could potentially be calculated using analytical prescription. Defaults to None, in which case Jx is calculated numerically as part of the propagation.
        :rtype Jx: array, optional
        :param Jx_diag: Bool to indicate whether the Jacobian matrix can be described with semi-diagonal elements. With this we mean that the measurand has the same shape as each of the input quantities and the square jacobain between the measurand and each of the input quantities individually, only has diagonal elements. Defaults to False
        :rtype Jx_diag: bool, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        if corr_x is None:
            corr_x = ["syst"] * len(x)
        for i in range(len(x)):
            if corr_x[i] is None:
                corr_x[i] = "syst"

        return self.propagate_standard(
            func,
            x,
            u_x,
            corr_x,
            param_fixed=param_fixed,
            corr_between=corr_between,
            return_corr=return_corr,
            return_Jacobian=return_Jacobian,
            repeat_dims=repeat_dims,
            corr_axis=corr_axis,
            fixed_corr_var=fixed_corr_var,
            output_vars=output_vars,
            Jx=Jx,
            Jx_diag=Jx_diag,
        )

    def propagate_cov(
        self,
        func,
        x,
        cov_x,
        param_fixed=None,
        corr_between=None,
        return_corr=False,
        return_Jacobian=False,
        repeat_dims=-99,
        corr_axis=-99,
        fixed_corr_var=False,
        output_vars=1,
        Jx=None,
        Jx_diag=None,
    ):
        """
        Propagate uncertainties with given covariance matrix through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        The covariance matrix can represent the full covariance matrix between all measurements in all dimensions.
        Alternatively if there are repeated measurements specified in repeat_dims, the covariance matrix is given
        for the covariance along the dimension that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param cov_x: list of covariance matrices on input quantities (usually numpy arrays). In case the input quantity is an array of shape (m,o), the covariance matrix needs to be given as an array of shape (m*o,m*o).
        :type cov_x: list[array]
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian matrix, defaults to False
        :type return_Jacobian: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param Jx: Jacobian matrix, evaluated at x. This allows to give a precomputed jacobian matrix, which could potentially be calculated using analytical prescription. Defaults to None, in which case Jx is calculated numerically as part of the propagation.
        :rtype Jx: array, optional
        :param Jx_diag: Bool to indicate whether the Jacobian matrix can be described with semi-diagonal elements. With this we mean that the measurand has the same shape as each of the input quantities and the square jacobain between the measurand and each of the input quantities individually, only has diagonal elements. Defaults to False
        :rtype Jx_diag: bool, optional
        :return: uncertainties on measurand
        :rtype: array
        """

        u_x = [cm.uncertainty_from_covariance(cov_x[i]) for i in range(len(x))]
        corr_x = [cm.correlation_from_covariance(cov_x[i]) for i in range(len(x))]

        return self.propagate_standard(
            func,
            x,
            u_x,
            corr_x,
            param_fixed=param_fixed,
            corr_between=corr_between,
            return_corr=return_corr,
            return_Jacobian=return_Jacobian,
            repeat_dims=repeat_dims,
            corr_axis=corr_axis,
            fixed_corr_var=fixed_corr_var,
            output_vars=output_vars,
            Jx=Jx,
            Jx_diag=Jx_diag,
        )

    def propagate_standard(
        self,
        func,
        x,
        u_x,
        corr_x,
        param_fixed=None,
        corr_between=None,
        return_corr=False,
        return_Jacobian=False,
        repeat_dims=-99,
        corr_axis=-99,
        fixed_corr_var=False,
        output_vars=1,
        Jx=None,
        Jx_diag=None,
    ):
        """
        Propagate uncertainties through measurement function with n input quantities. Correlations must be specified in corr_x.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Systematic uncertainties arise when there is full correlation between repeated measurements.
        There is a often also a correlation between measurements along the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array]
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian matrix, defaults to False
        :type return_Jacobian: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param Jx: Jacobian matrix, evaluated at x. This allows to give a precomputed jacobian matrix, which could potentially be calculated using analytical prescription. Defaults to None, in which case Jx is calculated numerically as part of the propagation.
        :rtype Jx: array, optional
        :param Jx_diag: Bool to indicate whether the Jacobian matrix can be described with semi-diagonal elements. With this we mean that the measurand has the same shape as each of the input quantities and the square jacobain between the measurand and each of the input quantities individually, only has diagonal elements. Defaults to False
        :rtype Jx_diag: bool, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        (
            fun,
            xflat,
            u_xflat,
            yshape,
            u_x,
            repeat_dims,
            n_repeats,
            repeat_shape,
            corr_axis,
            fixed_corr,
            Jx_diag,
        ) = self.perform_checks(
            func,
            x,
            u_x,
            corr_x,
            repeat_dims,
            corr_axis,
            output_vars,
            fixed_corr_var,
            Jx_diag,
            param_fixed,
        )

        if n_repeats > 0:
            inputs = np.empty(n_repeats, dtype=object)
            for i in range(n_repeats):
                xb, u_xb = util.select_repeated_x(
                    x, u_x, param_fixed, i, repeat_dims, repeat_shape
                )
                if Jx is not None:
                    Jxi = Jx[i]
                else:
                    Jxi = Jx
                inputs[i] = [
                    func,
                    xb,
                    u_xb,
                    corr_x,
                    None,
                    corr_between,
                    return_corr,
                    return_Jacobian,
                    -99,
                    corr_axis,
                    fixed_corr_var,
                    output_vars,
                    Jxi,
                    Jx_diag,
                ]

            if self.parallel_cores > 1:
                pool = Pool(self.parallel_cores)
                outs = pool.starmap(self.propagate_systematic, inputs)
                pool.close()

            else:
                outs = np.empty(n_repeats, dtype=object)
                for i in range(n_repeats):
                    outs[i] = self.propagate_standard(*inputs[i])

            return self.combine_repeated_outs(
                outs,
                yshape,
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_Jacobian,
                output_vars,
            )

        else:
            if Jx is None:
                Jx = cm.calculate_Jacobian(fun, xflat, Jx_diag, self.step)

            if corr_between is None:
                corr_between = np.eye(len(x))

            corrs = np.empty(len(x), dtype=object)
            for i in range(len(x)):
                if corr_x[i] == "syst":
                    corrs[i] = np.ones((len(x[i].ravel()), len(x[i].ravel())))
                elif corr_x[i] == "rand":
                    corrs[i] = np.eye(len(x[i].ravel()))
                else:
                    corrs[i] = corr_x[i]

            corr_x = cm.calculate_flattened_corr(corrs, corr_between)
            cov_x = cm.convert_corr_to_cov(corr_x, u_xflat)

            return self.process_jacobian(
                Jx,
                cov_x,
                yshape,
                return_corr,
                corr_axis,
                fixed_corr,
                output_vars,
                return_Jacobian,
            )

    def propagate_flattened_cov(
        self,
        func,
        x,
        flat_cov_x,
        return_corr=False,
        return_Jacobian=False,
        corr_axis=-99,
        output_vars=1,
        Jx=None,
        Jx_diag=None,
    ):
        """
        Propagate uncertainties with given covariance matrix through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        The covariance matrix can represent the full covariance matrix between all measurements in all dimensions.
        Alternatively if there are repeated measurements specified in repeat_dims, the covariance matrix is given
        for the covariance along the dimension that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param flat_cov_x: flattened covariance matrix on flattened (concatenated) input quantities (usually numpy arrays). In case there is n input quantity arrays of shape (m,o), the covariance matrix needs to be given as an array of shape (n*m*o,n*m*o).
        :type flat_cov_x: list[array]
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian matrix, defaults to False
        :type return_Jacobian: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param Jx: Jacobian matrix, evaluated at x. This allows to give a precomputed jacobian matrix, which could potentially be calculated using analytical prescription. Defaults to None, in which case Jx is calculated numerically as part of the propagation.
        :rtype Jx: array, optional
        :param Jx_diag: Bool to indicate whether the Jacobian matrix can be described with semi-diagonal elements. With this we mean that the measurand has the same shape as each of the input quantities and the square jacobain between the measurand and each of the input quantities individually, only has diagonal elements. Defaults to False
        :rtype Jx_diag: bool, optional
        :return: uncertainties on measurand
        :rtype: array
        """

        (
            fun,
            xflat,
            u_xflat,
            yshape,
            u_x,
            repeat_dims,
            n_repeats,
            repeat_shape,
            corr_axis,
            fixed_corr,
            Jx_diag,
        ) = self.perform_checks(
            func,
            x,
            None,
            None,
            -99,
            corr_axis,
            output_vars,
            False,
            Jx_diag,
            None,
        )

        if Jx is None:
            Jx = cm.calculate_Jacobian(fun, xflat, Jx_diag, self.step)

        return self.process_jacobian(
            Jx,
            flat_cov_x,
            yshape,
            return_corr,
            corr_axis,
            output_vars=output_vars,
            return_Jacobian=return_Jacobian,
        )

    def process_jacobian(
        self,
        J,
        covx,
        shape_y,
        return_corr,
        corr_axis,
        fixed_corr=None,
        output_vars=1,
        return_Jacobian=False,
    ):
        covy = np.dot(np.dot(J, covx), J.T)
        u_func = np.diag(covy) ** 0.5
        if fixed_corr is None:
            corr_y = cm.convert_cov_to_corr(covy, u_func)
        else:
            corr_y = fixed_corr
        if corr_axis >= 0:
            lencorr = shape_y[corr_axis] * output_vars
            corr_y = np.average(
                [
                    corr_y[
                        i * lencorr : (i + 1) * lencorr, i * lencorr : (i + 1) * lencorr
                    ]
                    for i in range(int(len(corr_y) / lencorr))
                ],
                axis=0,
            )

        if output_vars == 1:
            u_func = u_func.reshape(shape_y)
        else:
            u_func = u_func.reshape((output_vars,) + shape_y)

        if not return_corr:
            if not return_Jacobian:
                return u_func
            else:
                return u_func, J
        else:
            if output_vars == 1:
                if not return_Jacobian:
                    return u_func, corr_y
                else:
                    return u_func, corr_y, J
            else:
                # create an empty arrays and then populate it with the correlation matrix for each output parameter individually
                corr_ys, corr_out = cm.separate_flattened_corr(corr_y, output_vars)
                if not return_Jacobian:
                    return u_func, corr_ys, corr_out
                else:
                    return u_func, corr_ys, corr_out, J

    def perform_checks(
        self,
        func,
        x,
        u_x,
        corr_x,
        repeat_dims,
        corr_axis,
        output_vars,
        fixed_corr_var,
        Jx_diag,
        param_fixed,
    ):
        """
        Perform checks on the input parameters and set up the appropriate keywords for further processing

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of uncertainties/covariances on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :return: yshape,u_x,repeat_axis,repeat_dims,corr_axis,fixed_corr
        :rtype: tuple, list[array], int, int, int, array
        """

        xshapes = [xi.shape for xi in x]
        if (xshapes.count(xshapes[0]) == len(xshapes)) and repeat_dims == -99:
            if output_vars == 1:
                fun = lambda c: func(*c.reshape(len(x), -1))
            else:
                fun = lambda c: np.concatenate(func(*c.reshape(len(x), -1)))
        else:

            clims = [0]
            for i in range(len(x)):
                size = 1
                for dim in xshapes[i]:
                    size *= dim
                clims = np.append(clims, clims[i] + size)
            if output_vars == 1:
                fun = lambda c: func(
                    *[
                        c[clims[i] : clims[i + 1]].reshape(x[i].shape)
                        for i in range(len(x))
                    ]
                )
            else:
                fun = lambda c: np.concatenate(
                    func(
                        *[
                            c[clims[i] : clims[i + 1]].reshape(x[i].shape)
                            for i in range(len(x))
                        ]
                    )
                )

        # find the shape
        if output_vars == 1:
            yshape = np.array(func(*x)).shape
        else:
            yshape = np.array(func(*x)[0]).shape

        # shapewarning=False
        # for i in range(len(x)):
        #     if hasattr(x[i], "__len__"):
        #         if x[i].shape != yshape and self.parallel_cores == 0:
        #             shapewarning=True
        #     elif self.parallel_cores == 0:
        #         shapewarning=True
        #
        # if shapewarning:
        #     warnings.warn(
        #         "It looks like one of your input quantities is not an array or does not "
        #         "have the same shape as the measurand. This is not a problem, but means "
        #         "you likely cannot use array operations in your measurement function. "
        #         "You might need to set parallel_cores to 1 or higher when creating "
        #         "your MCPropagation object.")

        # Check for which input quantities there is no uncertainty,
        # replacing Nones with zeros where necessary.
        # Count the number of non-zero uncertainties. If this number is one, the
        # correlation matrix for the measurand will be the same as for this input qty.

        count = 0
        for i in range(len(x)):
            if u_x[i] is None:
                if hasattr(x[i], "__len__"):
                    u_x[i] = np.zeros(x[i].shape)
                else:
                    u_x[i] = 0.0
            if np.sum(u_x[i]) != 0 and fixed_corr_var == True:
                count += 1
                var = i
            if corr_x is not None:
                if corr_x[i] is not None:
                    if not isinstance(corr_x[i], str):
                        if np.any(corr_x[i] > 1.000001):
                            raise ValueError(
                                "One of the provided correlation matrices "
                                "has elements >1."
                            )

        if count == 1:
            fixed_corr_var = var
        else:
            fixed_corr_var = -99

        if fixed_corr_var >= 0 and corr_x is not None:
            if corr_x[fixed_corr_var] == "rand":
                fixed_corr = np.eye(len(u_x[fixed_corr_var]))
            elif corr_x[fixed_corr_var] == "syst":
                fixed_corr = np.ones(
                    (len(u_x[fixed_corr_var]), len(u_x[fixed_corr_var]))
                )
            else:
                fixed_corr = corr_x[fixed_corr_var]

        else:
            fixed_corr = None

        # Set up repeat_axis and repeat_dims for proper use in recursive function.
        if isinstance(repeat_dims, int):
            repeat_dims = [repeat_dims]

        if len(repeat_dims) == 1:
            if repeat_dims[0] >= 0:
                n_repeats = yshape[repeat_dims[0]]
                if corr_axis > repeat_dims[0]:
                    corr_axis -= 1
                elif corr_axis == repeat_dims[0]:
                    print("corr_axis and repeat_axis keywords should not be the same.")
                    exit()
            else:
                n_repeats = 0
                if param_fixed is not None:
                    print("param_fixed should not be set unless repeat_dims is set.")
                    exit()

            repeat_shape = (n_repeats,)
            # repeat_dims = -99
        else:
            repeat_dims = -np.sort(-np.array(repeat_dims))
            n_repeats = yshape[repeat_dims[0]] * yshape[repeat_dims[1]]
            repeat_shape = (yshape[repeat_dims[0]], yshape[repeat_dims[1]])
            if corr_axis > repeat_dims[0]:
                corr_axis -= 1
            elif corr_axis == repeat_dims[0]:
                print("corr_axis and repeat_axis keywords should not be the same.")
                exit()
            if corr_axis > repeat_dims[1]:
                corr_axis -= 1
            elif corr_axis == repeat_dims[1]:
                print("corr_axis and repeat_axis keywords should not be the same.")
                exit()

        xflat = np.concatenate([xi.ravel() for xi in x])
        u_xflat = np.concatenate([u_xi.ravel() for u_xi in u_x])

        if Jx_diag is None:
            Jx_diag = self.Jx_diag

        return (
            fun,
            xflat,
            u_xflat,
            yshape,
            u_x,
            repeat_dims,
            n_repeats,
            repeat_shape,
            corr_axis,
            fixed_corr,
            Jx_diag,
        )

    def combine_repeated_outs(
        self,
        outs,
        yshape,
        n_repeats,
        repeat_shape,
        repeat_dims,
        return_corr,
        return_Jacobian,
        output_vars,
    ):
        """
        Combine the outputs of the repeated measurements into one results array

        :param outs: list of outputs of the repeated measurements
        :type outs: list[array]
        :param yshape: shape of the measurand
        :type yshape: tuple
        :param n_repeats: number of repeated measurements
        :type n_repeats: int
        :param repeat_shape: shape along which the measurements are repeated
        :type repeat_shape: tuple
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian, defaults to False
        :type return_Jacobian: bool, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: combined outputs
        :rtype: list[array]
        """

        if not return_corr and output_vars == 1 and not return_Jacobian:
            u_func = np.array([outs[i] for i in range(n_repeats)])

        elif output_vars == 1:
            u_func = np.array([outs[i][0] for i in range(n_repeats)])

        else:
            u_func = np.array(
                [
                    [outs[i][0][ii] for i in range(n_repeats)]
                    for ii in range(output_vars)
                ]
            )

        if len(repeat_dims) == 1:
            if output_vars == 1:
                u_func = np.moveaxis(u_func, 0, repeat_dims[0])
            else:
                u_func = np.moveaxis(u_func, 1, repeat_dims[0] + 1)

        else:
            if output_vars == 1:
                u_func = u_func.reshape(repeat_shape + (-1,))
                u_func = np.moveaxis(u_func, 0, repeat_dims[0])
                u_func = np.moveaxis(u_func, 0, repeat_dims[1])
            else:
                u_func = u_func.reshape((output_vars,) + repeat_shape + (-1,))
                u_func = np.moveaxis(u_func, 1, repeat_dims[0] + 1)
                u_func = np.moveaxis(u_func, 1, repeat_dims[1] + 1)

        if (output_vars == 1 and u_func.shape != yshape) or (
            output_vars > 1 and u_func[0].shape != yshape
        ):
            # print(u_func.shape, yshape)
            raise ValueError(
                "The shape of the uncertainties does not match the shape"
                "of the measurand. This is likely a problem with combining"
                "repeated measurements (repeat_dims keyword)."
            )

        if not return_corr and not return_Jacobian:
            return u_func

        else:
            returns = np.empty(len(outs[0]), dtype=object)
            returns[0] = u_func
            extra_index = 0
            if return_corr:
                corr = np.mean([outs[i][1] for i in range(n_repeats)], axis=0)
                returns[1] = corr
                extra_index += 1

                if output_vars > 1:
                    corr_out = np.mean(
                        [outs[i][1 + extra_index] for i in range(n_repeats)], axis=0
                    )
                    returns[1 + extra_index] = corr_out
                    extra_index += 1

            if return_Jacobian:
                returns[1 + extra_index] = [
                    outs[i][1 + extra_index] for i in range(n_repeats)
                ]

            return returns



"""Use Monte Carlo to propagate uncertainties"""

import copy
import time
import warnings
from abc import ABC,abstractmethod

import numpy as np
import obsarray

from punpy.digital_effects_table.digital_effects_table_templates import (
    DigitalEffectsTableTemplates,)
from punpy.digital_effects_table.measurement_function_utils import (
    MeasurementFunctionUtils,)
from punpy.mc.mc_propagation import MCPropagation

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


class MeasurementFunction(ABC):
    def __init__(
        self,
        prop=None,
        xvariables=None,
        uncxvariables=None,
        yvariable=None,
        yunit="",
        corr_between=None,
        param_fixed=None,
        repeat_dims=-99,
        corr_dims=-99,
        allow_some_nans=True,
        ydims=None,
        refxvar=None,
        sizes_dict=None,
        use_err_corr_dict=False,
    ):
        """
        Initialise MeasurementFunction

        :param prop: punpy MC propagation or LPU propagation object. Defaults to None, in which case a MC propagation object with 100 MC steps is used.
        :type prop: punpy.MCPropagation or punpy. LPUPropagation
        :param xvariables: list of input quantity names, in same order as arguments in measurement function and with same exact names as provided in input datasets. Defaults to None, in which case get_argument_names function is used.
        :type xvariables: list(str), optional
        :param uncxvariables: list of input quantity names for which uncertainties should be propagated. Should be a subset of input quantity names. Defaults to None, in which case uncertainties on all input quantities are used.
        :type uncxvariables: list(str), optional
        :param yvariable: name of measurand. Defaults to None, in which case get_measurand_name_and_unit function is used.
        :type yvariable: str, optional
        :param yunit: unit of measurand. Defaults to "" (unitless).
        :type yunit: str, optional
        :param corr_between: Allows to specify the (average) error correlation coefficient between the various input quantities. Defaults to None, in which case no error-correlation is assumed.
        :type corr_between: numpy.ndarray , optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param repeat_dims: Used to select the axis which has repeated measurements. Axis can be specified using the name(s) of the dimension, or their index in the ydims. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: str or int or list(str) or list(int), optional
        :param corr_dims: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_dims: integer, optional
        :param ydims: list of dimensions of the measurand, in correct order. list of list of dimensions when there are multiple measurands. Default to None, in which case it is assumed to be the same as refxvar (see below) input quantity.
        :type ydims: list(str), optional
        :param refxvar: name of reference input quantity that has the same shape as measurand. Defaults to None
        :type refxvar: string, optional
        :param sizes_dict: Dictionary with sizes of each of the dimensions of the measurand. Defaults to None, in which cases sizes come from input quantites.
        :type sizes_dict: dict, optional
        :param use_err_corr_dict: when possible, use dictionaries with separate error-correlation info per dimension in order to save memory
        :type use_err_corr_dict: bool, optional
        """
        if prop is None:
            self.prop = MCPropagation(100, dtype="float32")

        else:
            self.prop = prop

        self.xvariables = None
        if self.get_argument_names() is None:
            self.xvariables = xvariables
        else:
            self.xvariables = self.get_argument_names()
            if xvariables is not None:
                if xvariables != self.xvariables:
                    raise ValueError(
                        "punpy.MeasurementFunction: when specifying both xvariables and get_argument_names, they need to be the same!"
                    )

        if self.xvariables is None:
            raise ValueError(
                "punpy.MeasurementFunction: You need to specify xvariables as keyword when initialising MeasurementFunction object, or add get_argument_names() as a function of the class."
            )

        if uncxvariables is None:
            self.uncxvariables = self.xvariables
        else:
            if isinstance(uncxvariables,str):
                uncxvariables=[uncxvariables]

            if np.all([uncvar in self.xvariables for uncvar in uncxvariables]):
                self.uncxvariables = uncxvariables

            else:
                raise ValueError(
                    "punpy.MeasurementFunction: the provided uncxvariables are not a subset of the provided xvariables."
                )

        if yvariable is None:
            self.yvariable, yunit = self.get_measurand_name_and_unit()
        else:
            self.yvariable = yvariable
            yvar = self.get_measurand_name_and_unit()[0]
            if yvar != "measurand":
                if yvariable != yvar:
                    raise ValueError(
                        "punpy.MeasurementFunction: when specifying both yvariable and get_measurand_name_and_unit, they need to be the same!"
                    )

        self.corr_between = corr_between
        if isinstance(self.yvariable,str):
            self.output_vars = 1
        else:
            self.output_vars = len(self.yvariable)
        self.ydims = ydims

        if self.output_vars==1:
            if isinstance(self.yvariable,str):
                self.yvariable=[self.yvariable]
            if isinstance(yunit,str):
                yunit=[yunit]
            if self.ydims is not None:
                if isinstance(self.ydims[0],str):
                    self.ydims = [ydims]

        self.templ = DigitalEffectsTableTemplates(self.yvariable, yunit, self.output_vars)
        self.sizes_dict = sizes_dict

        if refxvar is None:
            self.refxvar = self.xvariables[0]
        elif isinstance(refxvar, int):
            self.refxvar = self.xvariables[refxvar]
        else:
            self.refxvar = refxvar

        if isinstance(repeat_dims, int) or isinstance(repeat_dims, str):
            repeat_dims = [repeat_dims]

        self.repeat_dims = np.array(repeat_dims)

        self.num_repeat_dims = np.empty_like(self.repeat_dims, dtype=int)
        self.str_repeat_dims = np.empty_like(self.repeat_dims, dtype="<U30")

        if isinstance(corr_dims, int) or isinstance(corr_dims, str):
            corr_dims = [corr_dims]
        self.corr_dims = np.array(corr_dims,dtype=object)
        self.num_corr_dims = np.empty_like(self.corr_dims,dtype=object)
        self.str_corr_dims = np.empty_like(self.corr_dims,dtype="<U30")

        self.str_repeat_noncorr_dims = []

        self.param_fixed = param_fixed

        self.utils = MeasurementFunctionUtils(
            self.xvariables,
            self.uncxvariables,
            self.ydims,
            self.str_repeat_dims,
            self.str_repeat_noncorr_dims,
            self.prop.verbose,
            self.templ,
            use_err_corr_dict,
        )

        self.use_err_corr_dict = use_err_corr_dict
        self.allow_some_nans = allow_some_nans

    @abstractmethod
    def meas_function(self, *args, **kwargs):
        """
        meas_function is the measurement function itself, to be used in the uncertainty propagation.
        This function must be overwritten by the user when creating their MeasurementFunction subclass.
        """
        pass

    def get_argument_names(self):
        """
        This function allows to return the names of the input quantities as a list of strings.
        Can optionally be overwritten to provide names instead of providing xvariables as a keyword.

        :return: names of the input quantities
        :rtype: list of strings
        """
        return None

    def get_measurand_name_and_unit(self):
        """
        This function allows to return the name and unit of the measurand as strings.
        Can optionally be overwritten to provide names instead of providing yvariable as a keyword.

        :return: name of the measurand, unit
        :rtype: tuple(str, str)
        """
        return "measurand", ""

    def update_measurand(self, measurand, measurand_unit):
        self.yvariable = measurand
        self.templ = DigitalEffectsTableTemplates(self.yvariable, measurand_unit)

    def setup(self, *args, **kwargs):
        """
        This function is to provide a setup stage that can be run before propagating uncertainties.
        This allows to set up additional class attributes etc, without needing to edit the initialiser (which is quite specific to this class).
        This function can optionally be overwritten by the user when creating their MeasurementFunction subclass.
        """
        pass

    def propagate_ds(
        self,
        *args,
        store_unc_percent=False,
        expand=False,
        ds_out_pre=None,
        use_ds_out_pre_unmodified=False,
        include_corr=True,
    ):
        """
        Function to propagate the uncertainties on the input quantities present in the
        digital effects tables provided as the input arguments, through the measurement
        function to produce an output digital effects table with the combined random,
        systematic and structured uncertainties on the measurand

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param store_unc_percent: Boolean defining whether relative uncertainties should be returned or not. Default to True (relative uncertaintie returned)
        :type store_unc_percent: bool (optional)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :param ds_out_pre: Pre-existing output dataset in which the measurand & uncertainty variables should be saved. Defaults to None, in which case a new dataset is created.
        :type ds_out_pre: xarray.dataset (optional)
        :param include_corr: boolean to indicate whether the output dataset should include the correlation matrices. Defaults to True.
        :type include_corr: bool (optional)
        :return: digital effects table with uncertainties on measurand
        :rtype: obsarray dataset
        """
        if self.prop.verbose:
            print(
                "starting propagate_ds (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )

        # first calculate the measurand and propagate the uncertainties
        y = self.check_sizes_and_run(*args, expand=expand, ds_out_pre=ds_out_pre)

        u_rand_y = self.propagate_random(*args, expand=expand)
        if self.prop.verbose:
            print(
                "propagate_random done (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )

        u_syst_y = self.propagate_systematic(*args, expand=expand)
        if self.prop.verbose:
            print(
                "propagate systematic done (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )

        if include_corr:
            if self.output_vars==1:
                u_stru_y, corr_stru_y = self.propagate_structured(
                *args, expand=expand, return_corr=include_corr
                )

            else:
                u_stru_y, corr_stru_y, corr_stru_y_between = self.propagate_structured(
                    *args, expand=expand, return_corr=include_corr
                )
        else:
            u_stru_y = self.propagate_structured(
                *args, expand=expand, return_corr=include_corr
            )
            corr_stru_y = None

        if use_ds_out_pre_unmodified:
            if ds_out_pre is not None:
                ds_out=ds_out_pre
            else:
                raise ValueError(
                    "punpy.MeasurementFunction: ds_out_pre needs to be provided when use_ds_out_pre_unmodified is set to True."
                )
        else:
            repeat_dim_err_corrs = self.utils.find_repeat_dim_corr(
                "str", *args, store_unc_percent=store_unc_percent, ydims=self.ydims
            )

            self.utils.set_repeat_dims_form(repeat_dim_err_corrs)

            template = self.templ.make_template_main(
                self.ydims,
                self.sizes_dict,
                store_unc_percent=store_unc_percent,
                str_repeat_noncorr_dims=self.str_repeat_noncorr_dims,
                repeat_dim_err_corrs=repeat_dim_err_corrs,
            )

            # create dataset template
            ds_out = obsarray.create_ds(template, self.sizes_dict)

        #add trivial first dimension to so we can loop over output_vars later
        if self.output_vars==1:
            if u_rand_y is not None:
                u_rand_y=u_rand_y[None,...]
            if u_syst_y is not None:
                u_syst_y=u_syst_y[None,...]
            if u_stru_y is not None:
                u_stru_y=u_stru_y[None,...]
            if corr_stru_y is not None:
                corr_stru_y=corr_stru_y[None,...]

        for i in range(self.output_vars):
            ds_out[self.yvariable[i]].values = y[i]

            if store_unc_percent:
                ucomp_ran = "u_rel_ran_" + self.yvariable[i]
                ucomp_sys = "u_rel_sys_" + self.yvariable[i]
                ucomp_str = "u_rel_str_" + self.yvariable[i]
            else:
                ucomp_ran = "u_ran_" + self.yvariable[i]
                ucomp_sys = "u_sys_" + self.yvariable[i]
                ucomp_str = "u_str_" + self.yvariable[i]

            if u_rand_y is None:
                ds_out = self.templ.remove_unc_component(ds_out, self.yvariable[i], ucomp_ran)
            else:
                if store_unc_percent:
                    ds_out[ucomp_ran].values = u_rand_y[i] / y[i] * 100
                else:
                    ds_out[ucomp_ran].values = u_rand_y[i]

            if u_syst_y is None:
                ds_out = self.templ.remove_unc_component(ds_out, self.yvariable[i], ucomp_sys)
            else:
                if store_unc_percent:
                    ds_out[ucomp_sys].values = u_syst_y[i] / y[i] * 100
                else:
                    ds_out[ucomp_sys].values = u_syst_y[i]

            if u_stru_y is None:
                ds_out = self.templ.remove_unc_component(
                    ds_out,
                    self.yvariable[i],
                    ucomp_str,
                    err_corr_comp="err_corr_str_" + self.yvariable[i],
                )
            else:
                if store_unc_percent:
                    ds_out[ucomp_str].values = u_stru_y[i] / y[i] * 100
                else:
                    ds_out[ucomp_str].values = u_stru_y[i]

                if include_corr:
                    ds_out["err_corr_str_" + self.yvariable[i]].values = corr_stru_y[i]

                else:
                    ds_out.drop("err_corr_str_" + self.yvariable[i])

                    #ds_out.drop("err_corr_str_between")


            if (ds_out_pre is not None) and not use_ds_out_pre_unmodified:
                self.templ.join_with_preexisting_ds(ds_out, ds_out_pre, drop=self.yvariable)

        if self.prop.verbose:
            print(
                "finishing propagate_ds (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )

        return ds_out

    def propagate_ds_total(
        self,
        *args,
        store_unc_percent=False,
        expand=False,
        ds_out_pre=None,
        use_ds_out_pre_unmodified=False,
        include_corr=True,
    ):
        """
        Function to propagate the total uncertainties present in the digital effects
        tables in the input arguments, through the measurement function to produce
        an output digital effects table with the total uncertainties on the measurand

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param store_unc_percent: Boolean defining whether relative uncertainties should be returned or not. Default to True (relative uncertaintie returned)
        :type store_unc_percent: bool (optional)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :param ds_out_pre: Pre-existing output dataset in which the measurand & uncertainty variables should be saved. Defaults to None, in which case a new dataset is created.
        :type ds_out_pre: xarray.dataset (optional)
        :param include_corr: boolean to indicate whether the output dataset should include the correlation matrices. Defaults to True.
        :type include_corr: bool (optional)
        :return: digital effects table with uncertainties on measurand
        :rtype: obsarray dataset
        """
        if self.prop.verbose:
            print(
                "starting propagate_ds_total (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )
        y = self.check_sizes_and_run(*args, expand=expand, ds_out_pre=ds_out_pre)

        if include_corr:
            if self.output_vars==1:
                u_tot_y, corr_tot_y = self.propagate_total(
                    *args, expand=expand, return_corr=include_corr
                )
            else:
                u_tot_y, corr_tot_y, corr_tot_y_between = self.propagate_total(
                    *args, expand=expand, return_corr=include_corr
                )
        else:
            u_tot_y = self.propagate_total(
                *args, expand=expand, return_corr=include_corr
            )
            corr_tot_y = None

        if use_ds_out_pre_unmodified:
            if ds_out_pre is not None:
                ds_out=ds_out_pre
            else:
                raise ValueError(
                    "punpy.MeasurementFunction: ds_out_pre needs to be provided when use_ds_out_pre_unmodified is set to True."
                )
        else:
            repeat_dim_err_corrs = self.utils.find_repeat_dim_corr(
            "tot", *args, store_unc_percent=store_unc_percent, ydims=self.ydims
            )

            self.utils.set_repeat_dims_form(repeat_dim_err_corrs)

            template = self.templ.make_template_tot(
                self.ydims,
                self.sizes_dict,
                store_unc_percent=store_unc_percent,
                str_corr_dims=self.str_corr_dims,
                str_repeat_noncorr_dims=self.str_repeat_noncorr_dims,
                repeat_dim_err_corrs=repeat_dim_err_corrs,
            )

            # create dataset template
            ds_out = obsarray.create_ds(template, self.sizes_dict)

        if self.output_vars==1:
            u_tot_y=u_tot_y[None,...]
            corr_tot_y=corr_tot_y[None,...]

        for i in range(self.output_vars):
            ds_out[self.yvariable[i]].values = y[i]

            if store_unc_percent:
                ucomp = "u_rel_tot_" + self.yvariable[i]
            else:
                ucomp = "u_tot_" + self.yvariable[i]

            if u_tot_y is None:
                ds_out = self.templ.remove_unc_component(
                    ds_out,
                    self.yvariable[i],
                    ucomp,
                    err_corr_comp="err_corr_tot_" + self.yvariable[i],
                )
            else:
                if store_unc_percent:
                    ds_out[ucomp].values = u_tot_y[i] / y[i] * 100
                else:
                    ds_out[ucomp].values = u_tot_y[i]

                if include_corr:
                    if len(self.str_corr_dims)==1:
                        ds_out["err_corr_tot_" + self.yvariable[i]].values = corr_tot_y[i]
                    else:
                        for ii in range(len(self.str_corr_dims)):
                            ds_out["err_corr_tot_" + self.yvariable[i]+"_"+self.str_corr_dims[ii]].values = corr_tot_y[i][ii]

                else:
                    if len(self.str_corr_dims)==1:
                        ds_out.drop("err_corr_tot_" + self.yvariable[i])
                    else:
                        for ii in range(len(self.str_corr_dims)):
                            ds_out.drop("err_corr_tot_" + self.yvariable[i]+"_"+self.str_corr_dims[ii])

            if (ds_out_pre is not None) and not use_ds_out_pre_unmodified:
                self.templ.join_with_preexisting_ds(ds_out, ds_out_pre, drop=self.yvariable[i])

        #ds_out["err_corr_tot_between"]= corr_tot_y_between
        #ds_out.drop("err_corr_tot_between")

        if self.prop.verbose:
            print(
                "finishing propagate_ds_total (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )

        return ds_out

    def propagate_ds_specific(
        self,
        comp_list,
        *args,
        comp_list_out=None,
        store_unc_percent=False,
        expand=False,
        ds_out_pre=None,
            use_ds_out_pre_unmodified=False,
        include_corr=True,
        simple_random=True,
        simple_systematic=True,
    ):
        """
        Function to propagate the uncertainties on the input quantities present in the
        digital effects tables provided as the input arguments, through the measurement
        function to produce an output digital effects table with the uncertainties of specific
        components listed in comp_list.

        :param comp_list: list of uncertainty contributions to propagate
        :rtype comp_list: list of strings or string
        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param store_unc_percent: Boolean defining whether relative uncertainties should be returned or not. Default to True (relative uncertaintie returned)
        :type store_unc_percent: bool (optional)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :param ds_out_pre: Pre-existing output dataset in which the measurand & uncertainty variables should be saved. Defaults to None, in which case a new dataset is created.
        :type ds_out_pre: xarray.dataset (optional)
        :param include_corr: boolean to indicate whether the output dataset should include the correlation matrices. Defaults to True.
        :type include_corr: bool (optional)
        :return: digital effects table with uncertainties on measurand
        :rtype: obsarray dataset
        """
        if self.prop.verbose:
            print(
                "starting propagate_ds_specific (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )

        if isinstance(comp_list, str):
            comp_list = [comp_list]

        if comp_list_out is None:
            comp_list_out = comp_list

        # first calculate the measurand and propagate the uncertainties
        y = self.check_sizes_and_run(*args, expand=expand, ds_out_pre=ds_out_pre)

        if use_ds_out_pre_unmodified:
            if ds_out_pre is not None:
                ds_out=ds_out_pre
            else:
                raise ValueError(
                    "punpy.MeasurementFunction: ds_out_pre needs to be provided when use_ds_out_pre_unmodified is set to True."
                )
        else:
            repeat_dim_err_corrs = [
                self.utils.find_repeat_dim_corr(
                form, *args, store_unc_percent=store_unc_percent, ydims=self.ydims
                )
                for form in comp_list
            ]

            self.utils.set_repeat_dims_form(repeat_dim_err_corrs)

            template = self.templ.make_template_specific(
                comp_list_out,
                self.ydims,
                self.sizes_dict,
                store_unc_percent=store_unc_percent,
                str_repeat_noncorr_dims=self.str_repeat_noncorr_dims,
                repeat_dim_err_corrs=repeat_dim_err_corrs,
                simple_random=simple_random,
                simple_systematic=simple_systematic,
            )

            # create dataset template
            ds_out = obsarray.create_ds(template, self.sizes_dict)

        for i in range(self.output_vars):
            ds_out[self.yvariable[i]].values = y[i]

        for icomp, comp in enumerate(comp_list):
            if comp == "random":
                u_comp_y = self.propagate_random(*args, expand=expand)

            elif comp == "systematic":
                u_comp_y = self.propagate_systematic(*args, expand=expand)

            else:
                if include_corr:
                    if self.output_vars==1:
                        u_comp_y, corr_comp_y = self.propagate_specific(
                        comp, *args, return_corr=include_corr, expand=expand
                     )
                        u_comp_y=u_comp_y[None,...]
                        corr_comp_y=corr_comp_y[None,...]

                    else:
                        u_comp_y, corr_comp_y,corr_comp_y_between = self.propagate_specific(
                            comp, *args, return_corr=include_corr, expand=expand
                        )
                else:
                    u_comp_y = self.propagate_specific(
                        comp, *args, return_corr=include_corr, expand=expand
                    )
                    corr_comp_y = None

                for i in range(self.output_vars):
                    if corr_comp_y is not None:
                        ds_out[
                            "err_corr_" + comp_list_out[icomp] + "_" + self.yvariable[i]
                        ].values = corr_comp_y[i]
                    else:
                        ds_out.drop(
                            "err_corr_" + comp_list_out[icomp] + "_" + self.yvariable[i]
                        )

                    if u_comp_y is None:
                        if store_unc_percent:
                            ds_out = self.templ.remove_unc_component(
                                ds_out,
                                self.yvariable[i],
                                "u_rel_" + comp_list_out[icomp] + "_" + self.yvariable[i],
                                )
                        else:
                            ds_out = self.templ.remove_unc_component(
                                ds_out,
                                self.yvariable[i],
                                "u_" + comp_list_out[icomp] + "_" + self.yvariable[i],
                                )
                    else:
                        if store_unc_percent:
                            ds_out[
                                "u_rel_" + comp_list_out[icomp] + "_" + self.yvariable[i]
                                ].values = (u_comp_y[i] / y[i] * 100)
                        else:
                            ds_out[
                                "u_" + comp_list_out[icomp] + "_" + self.yvariable[i]
                                ].values = u_comp_y[i]

        for i in range(self.output_vars):
            if (ds_out_pre is not None) and not use_ds_out_pre_unmodified:
                ds_out = self.templ.join_with_preexisting_ds(
                    ds_out, ds_out_pre, drop=self.yvariable[i]
                )

        if self.prop.verbose:
            print(
                "finishing propagate_ds_specific (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )

        return ds_out

    def propagate_ds_all(
        self,
        *args,
        store_unc_percent=False,
        expand=False,
        ds_out_pre=None,
        use_ds_out_pre_unmodified=False,
        include_corr=True,
    ):
        """
        Function to propagate the uncertainties on the input quantities present in the
        digital effects tables provided as the input arguments, through the measurement
        function to produce an output digital effects table with the combined random,
        systematic and structured uncertainties on the measurand

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param store_unc_percent: Boolean defining whether relative uncertainties should be returned or not. Default to True (relative uncertaintie returned)
        :type store_unc_percent: bool (optional)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :param ds_out_pre: Pre-existing output dataset in which the measurand & uncertainty variables should be saved. Defaults to None, in which case a new dataset is created.
        :type ds_out_pre: xarray.dataset (optional)
        :param include_corr: boolean to indicate whether the output dataset should include the correlation matrices. Defaults to True.
        :type include_corr: bool (optional)
        :return: digital effects table with uncertainties on measurand
        :rtype: obsarray dataset
        """

        comp_list = []
        for iv, var in enumerate(self.xvariables):
            for dataset in args:
                if var in dataset.keys():
                    comps = self.utils.find_comps("tot", dataset, var)
                    if comps is not None:
                        for comp in comps:
                            if isinstance(comp, str):
                                comp_name = comp
                            else:
                                comp_name = comp.name
                            comp_name = comp_name.replace("_" + var, "")
                            comp_name = comp_name.replace("u_rel_", "")
                            if comp_name[0:2] == "u_":
                                comp_name = comp_name[2::]
                            comp_list.append(comp_name)
        comp_list = np.unique(np.array(comp_list))
        return self.propagate_ds_specific(
            comp_list,
            *args,
            store_unc_percent=store_unc_percent,
            expand=expand,
            ds_out_pre=ds_out_pre,
            use_ds_out_pre_unmodified=use_ds_out_pre_unmodified,
            include_corr=include_corr,
        )

    def run(self, *args, expand=False):
        """
        Function to calculate the measurand by running input quantities through measurement function.

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :return: measurand
        :rtype: numpy.ndarray
        """
        input_qty = self.utils.get_input_qty(
            args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims
        )

        return np.array(self.meas_function(*input_qty))

    def check_sizes_and_run(self, *args, expand=False, ds_out_pre=None):
        """
        Function to check the sizes of the input quantities and measurand and perform some checks and preprocessing

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :return: None
        :rtype: None
        """
        if self.ydims is None:
            self.ydims = np.empty(self.output_vars,dtype=object)
            for i in range(self.output_vars):
                if ds_out_pre is not None:
                    self.ydims[i] = ds_out_pre[self.yvariable[i]].dims
                else:
                    for dataset in args:
                        try:
                            self.ydims[i] = dataset[self.refxvar].dims
                        except:
                            continue

        y = self.run(*args, expand=expand)

        if self.output_vars==1:
            y=y[None,...]

        if self.sizes_dict is None:
            self.sizes_dict = {}
            for i in range(self.output_vars):
                for idim, dim in enumerate(self.ydims[i]):
                    try:
                        self.sizes_dict[dim] = y[i].shape[idim]
                    except:
                        raise ValueError()

        for idimr in range(len(self.repeat_dims)):
            if isinstance(self.repeat_dims[idimr], str):
                self.str_repeat_dims[idimr] , self.num_repeat_dims[idimr] = self.check_and_convert_str_dims(self.repeat_dims[idimr])

            elif isinstance(self.repeat_dims[idimr], (int, np.integer)):
                if self.repeat_dims[idimr] >= 0:
                    self.str_repeat_dims[idimr] , self.num_repeat_dims[idimr] = self.check_and_convert_num_dims(self.repeat_dims[idimr])
                else:
                    self.num_repeat_dims[idimr]=self.repeat_dims[idimr]

            else:
                raise ValueError(
                    "punpy.measurment_function: repeat_dims needs to be provided as ints or strings"
                )

        str_repeat_noncorr_dims=[str_dim for str_dim in self.str_repeat_dims if str_dim!='']

        all_corr_dims=[]
        for idimc in range(len(self.corr_dims)):
            if isinstance(self.corr_dims[idimc], str):
                if "." in self.corr_dims[idimc]:
                    corr_dims=self.corr_dims[idimc].split(".")
                    str_corr_dims=np.empty_like(corr_dims)
                    num_corr_dims=np.empty_like(corr_dims,dtype=int)
                    if corr_dims[0].isdigit():
                        corrlen=1
                        for ic in range(len(corr_dims)):
                            str_corr_dims[ic], num_corr_dims[ic] = self.check_and_convert_num_dims(int(corr_dims[ic]))
                            all_corr_dims.append(str_corr_dims[ic])
                            corrlen*=y[0].shape[num_corr_dims[ic]]
                        self.num_corr_dims[idimc] = copy.copy(self.corr_dims[idimc])
                        self.str_corr_dims[idimc] = ".".join(str_corr_dims)
                        self.sizes_dict[self.str_corr_dims[idimc]]=corrlen

                    else:
                        corrlen=1
                        for ic in range(len(corr_dims)):
                            str_corr_dims[ic], num_corr_dims[ic] = self.check_and_convert_str_dims(corr_dims[ic])
                            all_corr_dims.append(str_corr_dims[ic])
                            corrlen*=y[0].shape[num_corr_dims[ic]]
                        self.str_corr_dims[idimc] = copy.copy(self.corr_dims[idimc])
                        self.num_corr_dims[idimc] = ".".join([str(cdim) for cdim in num_corr_dims])
                        self.sizes_dict[self.str_corr_dims[idimc]]=corrlen

                elif not self.corr_dims[idimc] in self.ydims[0]:
                    raise ValueError(
                        "punpy.measurement_function: The corr_dim (%s) is not in the measurand dimensions (%s)."
                        % (self.corr_dims[idimc], self.ydims)
                    )

                else:
                    self.str_corr_dims[idimc] , self.num_corr_dims[idimc] = self.check_and_convert_str_dims(self.corr_dims[idimc])
                    all_corr_dims.append(self.corr_dims[idimc])

            elif isinstance(self.corr_dims[idimc], (int, np.integer)):
                if self.corr_dims[idimc]>=0:
                    self.str_corr_dims[idimc], self.num_corr_dims[idimc] = self.check_and_convert_num_dims(self.corr_dims[idimc])
                    all_corr_dims.append(self.ydims[0][self.corr_dims[idimc]])
                else:
                    self.num_corr_dims[idimc]=self.corr_dims[idimc]
                    self.str_corr_dims=[]
                    all_corr_dims=self.ydims[0]
            else:
                raise ValueError(
                    "punpy.measurment_function: corr_dims needs to be provided as ints or strings"
                )


        for i in range(self.output_vars):
            for idim, dim in enumerate(self.ydims[i]):
                if dim not in all_corr_dims:
                    str_repeat_noncorr_dims.append(dim)

        self.str_repeat_noncorr_dims = np.array(str_repeat_noncorr_dims).flatten()

        for i in range(self.output_vars):
            #set sizes dict for combined shape of error correlation dict
            key=""
            val=1
            for idim, dim in enumerate(self.ydims[i]):
                if dim not in self.str_repeat_noncorr_dims:
                    key+="."+dim
                    val*=self.sizes_dict[dim]
            self.sizes_dict[key[1::]]=val


        if (not expand) and (self.param_fixed is None):
            self.param_fixed = [False] * len(self.xvariables)
            for iv, var in enumerate(self.xvariables):
                found = False
                for dataset in args:
                    if hasattr(dataset, "variables"):
                        if var in dataset.variables:
                            if all(
                                [
                                    self.str_repeat_dims[i] in dataset[var].dims
                                    for i in range(len(self.str_repeat_dims))
                                ]
                            ):
                                found = True

                if not found:
                    self.param_fixed[iv] = True
                    if self.prop.verbose:
                        print(
                            "Variable %s not found in repeat_dims. setting param_fixed to True"
                            % (var)
                        )

        self.utils.ydims = self.ydims
        self.utils.str_repeat_noncorr_dims = self.str_repeat_noncorr_dims
        self.utils.str_repeat_dims = self.str_repeat_dims

        return y

    def check_and_convert_str_dims(self,dim):
        for i in range(self.output_vars):
            if not dim in self.ydims[i]:
                raise ValueError(
                    "punpy.measurement_function: The repeat_dim (%s) is not in the measurand dimensions (%s)."
                    % (dim, self.ydims)
                )
        if not np.all([ydims.index(dim)==self.ydims[0].index(dim) for ydims in self.ydims]):
            warnings.warn(
                "punpy.measurement_function: The repeat_dim (%s) cannot be used because it is not at the same index for every ydims of every measurand (%s)."
                % (dim, self.ydims)
            )
        return copy.copy(dim),copy.copy(self.ydims[0].index(dim))

    def check_and_convert_num_dims(self,dim):
        if dim>=0:
            if not np.all([ydims[dim]==self.ydims[0][dim] for ydims in self.ydims]):
                warnings.warn(
                    "punpy.measurement_function: The repeat_dim (%s) cannot be used because it is not at the same index for every ydims of every measurand (%s)."
                    % (dim, self.ydims)
                )

        return copy.copy(self.ydims[0][dim]),dim

    def propagate_total(self, *args, expand=False, return_corr=True):
        """
        Function to propagate uncertainties for the total uncertainty component.

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :param return_corr:  boolean to indicate whether the measurand error-correlation matrices should be returned. Defaults to True.
        :type return_corr: bool (optional)
        :return: uncertainty on measurand for total uncertainty component, error-correlation matrix of measurand for total uncertainty component
        :rtype: tuple(numpy.ndarray, numpy.ndarray)
        """
        y = self.check_sizes_and_run(*args, expand=expand)
        input_qty = self.utils.get_input_qty(
            args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims
        )
        input_unc = self.utils.get_input_unc(
            "tot", args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims
        )
        input_corr = self.utils.get_input_corr(
            "tot", args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims
        )
        if self.prop.verbose:
            print(
                "inputs extracted (%s s since creation of prop object)"
                % (time.time() - self.prop.starttime)
            )
        if all([iu is None for iu in input_unc]):
            return None, None
        else:
            if self.use_err_corr_dict:
                MC_x = self.prop.generate_MC_sample(
                    input_qty,
                    input_unc,
                    corr_x=input_corr,
                    corr_between=self.corr_between,
                    comp_list=True,
                )
                MC_y = self.prop.run_samples(
                    self.meas_function, MC_x, output_vars=self.output_vars
                )
                return self.prop.process_samples(
                    MC_x,
                    MC_y,
                    return_corr=return_corr,
                    corr_dims=self.num_corr_dims,
                    output_vars=self.output_vars,
                )
            else:
                return self.prop.propagate_standard(
                    self.meas_function,
                    input_qty,
                    input_unc,
                    input_corr,
                    param_fixed=self.param_fixed,
                    corr_between=self.corr_between,
                    return_corr=return_corr,
                    return_samples=False,
                    repeat_dims=self.num_repeat_dims,
                    corr_dims=self.num_corr_dims,
                    output_vars=self.output_vars,
                )

    def propagate_random(self, *args, expand=False):
        """
        Function to propagate uncertainties for the random uncertainty component.

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :return: uncertainty on measurand for random uncertainty component
        :rtype: numpy.ndarray
        """
        y = self.check_sizes_and_run(*args, expand=expand)
        input_qty = self.utils.get_input_qty(
            args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        input_unc = self.utils.get_input_unc(
            "rand", args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        if all([iu is None for iu in input_unc]):
            return None

        else:
            return self.prop.propagate_random(
                self.meas_function,
                input_qty,
                input_unc,
                param_fixed=self.param_fixed,
                corr_between=self.corr_between,
                return_corr=False,
                return_samples=False,
                repeat_dims=self.num_repeat_dims,
                corr_dims=self.num_corr_dims,
                output_vars=self.output_vars,
            )

    def propagate_systematic(self, *args, expand=False):
        """
        Function to propagate uncertainties for the systemtic uncertainty component.

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :return: uncertainty on measurand for systematic uncertainty component
        :rtype: numpy.ndarray
        """
        y = self.check_sizes_and_run(*args, expand=expand)
        input_qty = self.utils.get_input_qty(
            args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        input_unc = self.utils.get_input_unc(
            "syst", args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        if all([iu is None for iu in input_unc]):
            return None
        else:
            return self.prop.propagate_systematic(
                self.meas_function,
                input_qty,
                input_unc,
                param_fixed=self.param_fixed,
                corr_between=self.corr_between,
                return_corr=False,
                return_samples=False,
                repeat_dims=self.num_repeat_dims,
                corr_dims=self.num_corr_dims,
                output_vars=self.output_vars,
            )

    def propagate_structured(self, *args, expand=False, return_corr=True):
        """
        Function to propagate uncertainties for the structured uncertainty component.

        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :param return_corr:  boolean to indicate whether the measurand error-correlation matrices should be returned. Defaults to True.
        :type return_corr: bool (optional)
        :return: uncertainty on measurand for structured uncertainty component, error-correlation matrix of measurand for structured uncertainty component
        :rtype: tuple(numpy.ndarray, numpy.ndarray)
        """
        y = self.check_sizes_and_run(*args, expand=expand)
        input_qty = self.utils.get_input_qty(
            args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        input_unc = self.utils.get_input_unc(
            "stru", args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        input_corr = self.utils.get_input_corr(
            "stru", args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )

        if all([iu is None for iu in input_unc]):
            return None, None
        else:
            if self.use_err_corr_dict:
                MC_x = self.prop.generate_MC_sample(
                    input_qty,
                    input_unc,
                    corr_x=input_corr,
                    corr_between=self.corr_between,
                    comp_list=True,
                )
                MC_y = self.prop.run_samples(
                    self.meas_function, MC_x, output_vars=self.output_vars
                )
                return self.prop.process_samples(
                    MC_x,
                    MC_y,
                    return_corr=return_corr,
                    corr_dims=self.num_corr_dims,
                    output_vars=self.output_vars,
                )
            else:
                return self.prop.propagate_standard(
                    self.meas_function,
                    input_qty,
                    input_unc,
                    input_corr,
                    param_fixed=self.param_fixed,
                    corr_between=self.corr_between,
                    return_corr=return_corr,
                    return_samples=False,
                    repeat_dims=self.num_repeat_dims,
                    corr_dims=self.num_corr_dims,
                    output_vars=self.output_vars,
                )

    def propagate_specific(self, form, *args, expand=False, return_corr=False):
        """
        Function to propagate uncertainties for a specific uncertainty component.

        :param form: name or type of uncertainty component
        :type form: str
        :param args: One or multiple digital effects tables with input quantities, defined with obsarray
        :type args: obsarray dataset(s)
        :param expand: boolean to indicate whether the input quantities should be expanded/broadcasted to the shape of the measurand. Defaults to False.
        :type expand: bool (optional)
        :param return_corr:  boolean to indicate whether the measurand error-correlation matrices should be returned. Defaults to True.
        :type return_corr: bool (optional)
        :return: uncertainty on measurand for specific uncertainty component, error-correlation matrix of measurand for specific uncertainty component
        :rtype: tuple(numpy.ndarray, numpy.ndarray)
        """
        y = self.check_sizes_and_run(*args, expand=expand)

        input_qty = self.utils.get_input_qty(
            args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        input_unc = self.utils.get_input_unc(
            form, args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        input_corr = self.utils.get_input_corr(
            form, args, expand=expand, sizes_dict=self.sizes_dict, ydims=self.ydims[0]
        )
        if all([iu is None for iu in input_unc]):
            return None, None
        else:
            if self.use_err_corr_dict:
                MC_x = self.prop.generate_MC_sample(
                    input_qty,
                    input_unc,
                    corr_x=input_corr,
                    corr_between=self.corr_between,
                )
                MC_y = self.prop.run_samples(
                    self.meas_function, MC_x, output_vars=self.output_vars
                )
                return self.prop.process_samples(
                    MC_x,
                    MC_y,
                    return_corr=return_corr,
                    corr_dims=self.num_corr_dims,
                    output_vars=self.output_vars,
                )
            else:
                return self.prop.propagate_standard(
                    self.meas_function,
                    input_qty,
                    input_unc,
                    input_corr,
                    param_fixed=self.param_fixed,
                    corr_between=self.corr_between,
                    return_corr=return_corr,
                    return_samples=False,
                    repeat_dims=self.num_repeat_dims,
                    corr_dims=self.num_corr_dims,
                    output_vars=self.output_vars,
                    allow_some_nans=self.allow_some_nans
                )



import obsarray
import xarray as xr
from variables import L0_IRR_VARIABLES, CAL_VARIABLES

ds_in = xr.open_dataset("test_l0.nc")  # read digital effects table
ds_cal = xr.open_dataset("test_cal.nc")  # read digital effects table

# define dim_size_dict to specify size of arrays
dim_sizes = {
    "wavelength": len(ds_in.wavelength),
    "scan": len(ds_in.scan),
}

dim_sizes_cal = {
    "wavelength": len(ds_cal.wavelength),
    "calibrationdates": 1,
    "nonlinearcoef": 13,
    "nonlineardates": 1,
    "wavcoef": 5,
    "wavdates": 1,
}

# create dataset template
ds_out = obsarray.create_ds(L0_IRR_VARIABLES, dim_sizes)
ds_cal_out = obsarray.create_ds(CAL_VARIABLES, dim_sizes_cal)

for key in ds_out.keys():
    ds_out[key].values = ds_in[key].values
ds_out.assign_coords(wavelength=ds_in.wavelength)
ds_out.assign_coords(scan=ds_in.scan)

for key in ds_cal_out.keys():
    ds_cal_out[key].values = ds_cal[key].values.reshape(ds_cal_out[key].shape)

# store example file
ds_out.to_netcdf("det_hypernets_l0.nc")
ds_cal_out.to_netcdf("det_hypernets_cal.nc")



"""
Tests for mc propagation class
"""
import os.path
import unittest

import numpy as np
import numpy.testing as npt
import xarray as xr
from punpy import MeasurementFunction,MCPropagation

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "28/07/2021"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


class HypernetsMF(MeasurementFunction):
    def setup(self, min_value):
        self.min_value = min_value

    def meas_function(self, digital_number, gains, dark_signal, non_linear, int_time):
        DN = digital_number - dark_signal
        DN[DN == 0] = self.min_value
        corrected_DN = DN / (
            non_linear[0]
            + non_linear[1] * DN
            + non_linear[2] * DN ** 2
            + non_linear[3] * DN ** 3
            + non_linear[4] * DN ** 4
            + non_linear[5] * DN ** 5
            + non_linear[6] * DN ** 6
            + non_linear[7] * DN ** 7
        )
        if gains.ndim == 1:
            return gains[:, None] * corrected_DN / int_time * 1000
        else:
            return gains * corrected_DN / int_time * 1000

    def get_argument_names(self):
        return [
            "digital_number",
            "gains",
            "dark_signal",
            "non_linearity_coefficients",
            "integration_time",
        ]


dir_path = os.path.dirname(os.path.realpath(__file__))
calib_data = xr.open_dataset(os.path.join(dir_path, "det_hypernets_cal.nc"))
L0data = xr.open_dataset(os.path.join(dir_path, "det_hypernets_l0.nc"))
L1data = xr.open_dataset(os.path.join(dir_path, "test_l1.nc"))


# Define your measurement function inside a subclass of MeasurementFunction
class IdealGasLaw(MeasurementFunction):
    @staticmethod
    def meas_function(pres, temp, n):
        return (n * temp * 8.134) / pres

# Define your measurement function inside a subclass of MeasurementFunction
class IdealGasLaw_2out(MeasurementFunction):
    def meas_function(self, pres, temp, n):
        return (n * temp * 8.134) / pres, pres/temp


dir_path = os.path.dirname(os.path.realpath(__file__))
ds = xr.open_dataset(os.path.join(dir_path, "digital_effects_table_gaslaw_example.nc"))

volume = np.ones(ds["temperature"].values.shape) * 0.9533
u_tot_volume = (
    np.ones(ds["temperature"].values.shape)
    * 0.9533
    * ((1 / 10000) ** 2 + (1 / 293) ** 2 + (0.4 / 293) ** 2 + (1 / 40) ** 2) ** 0.5
)

u_ran_volume = (
    np.ones(ds["temperature"].values.shape)
    * 0.9533
    * ((1 / 293) ** 2 + (1 / 40) ** 2) ** 0.5
)
u_sys_volume = np.ones(ds["temperature"].values.shape) * 0.9533 * (0.4 / 293)
u_str_volume = np.ones(ds["temperature"].values.shape) * 0.9533 * (1 / 10000)


class TestMeasurementFunction(unittest.TestCase):
    """
    Class for unit tests
    """

    def test_gaslaw(self):

        prop = MCPropagation(1000, dtype="float32", verbose=False, parallel_cores=4)

        gl = IdealGasLaw(
            prop, ["pressure", "temperature", "n_moles"], yvariable="volume", yunit="m^3"
        )
        ds_y_tot = gl.propagate_ds_total(ds)

        npt.assert_(ds_y_tot["u_tot_volume"].attrs["err_corr_1_params"][0] in list(ds_y_tot.variables))

        npt.assert_allclose(ds_y_tot["volume"].values, volume, rtol=0.002)

        npt.assert_allclose(ds_y_tot["u_tot_volume"].values, u_tot_volume, rtol=0.1)

        prop = MCPropagation(3000, dtype="float32", verbose=False)
        gl = IdealGasLaw(
            prop,
            ["pressure", "temperature", "n_moles"],
            yvariable="volume",
            yunit="m^3",
            repeat_dims=[0, 2],
        )
        ds_y = gl.propagate_ds(ds)

        npt.assert_(ds_y["u_str_volume"].attrs["err_corr_3_params"][0] in list(ds_y.variables))

        npt.assert_allclose(ds_y["volume"].values, volume, rtol=0.002)
        npt.assert_allclose(ds_y["u_ran_volume"].values, u_ran_volume, rtol=0.08)
        npt.assert_allclose(ds_y["u_sys_volume"].values, u_sys_volume, rtol=0.08)
        npt.assert_allclose(ds_y["u_str_volume"].values, u_str_volume, rtol=0.08)
        npt.assert_allclose(ds_y["u_str_volume"].values, u_str_volume, rtol=0.08)
        npt.assert_allclose(
            ds_y.unc["volume"].total_unc().values, u_tot_volume, rtol=0.12
        )

        gl = IdealGasLaw(
            prop,
            ["pressure", "temperature", "n_moles"],
            uncxvariables=["pressure"],
            yvariable="volume",
            yunit="m^3",
            repeat_dims=[0, 2],
        )
        ds_y = gl.propagate_ds(ds)

        npt.assert_allclose(ds_y["volume"].values, volume, rtol=0.002)

    def test_gaslaw_2out(self):
        prop = MCPropagation(1000, dtype="float32", verbose=False)

        gl = IdealGasLaw_2out(
            prop, ["pressure", "temperature", "n_moles"], yvariable=["volume","P/T"], yunit=["m^3","Pa/K"]
        )
        ds_y_tot = gl.propagate_ds_total(ds)

        npt.assert_allclose(ds_y_tot["volume"].values, volume, rtol=0.002)

        npt.assert_allclose(ds_y_tot["u_tot_volume"].values, u_tot_volume, rtol=0.1)

    def test_gaslaw_errcorrdict(self):
        prop = MCPropagation(1000, dtype="float32", verbose=False)

        gl = IdealGasLaw(
            prop,
            ["pressure", "temperature", "n_moles"],
            yvariable="volume",
            yunit="m^3",
            use_err_corr_dict=True,
        )
        ds_y_tot = gl.propagate_ds_total(ds)

        npt.assert_allclose(ds_y_tot["volume"].values, volume, rtol=0.002)

        npt.assert_allclose(ds_y_tot["u_tot_volume"].values, u_tot_volume, rtol=0.1)

        prop = MCPropagation(3000, dtype="float32", verbose=False)
        gl = IdealGasLaw(
            prop,
            ["pressure", "temperature", "n_moles"],
            yvariable="volume",
            yunit="m^3",
            repeat_dims=[0, 2],
        )
        ds_y = gl.propagate_ds(ds)

        npt.assert_allclose(ds_y["volume"].values, volume, rtol=0.002)
        npt.assert_allclose(ds_y["u_ran_volume"].values, u_ran_volume, rtol=0.06)
        npt.assert_allclose(ds_y["u_sys_volume"].values, u_sys_volume, rtol=0.06)
        npt.assert_allclose(ds_y["u_str_volume"].values, u_str_volume, rtol=0.06)
        npt.assert_allclose(
            ds_y.unc["volume"].total_unc().values, u_tot_volume, rtol=0.06
        )

    def test_gaslaw_corrdim(self):
        prop = MCPropagation(1000, dtype="float32", verbose=False)

        gl = IdealGasLaw(
            prop,
            ["pressure", "temperature", "n_moles"],
            yvariable="volume",
            yunit="m^3",
            corr_dims=["0.1",2],
        )
        ds_y_tot = gl.propagate_ds_total(ds)

        npt.assert_allclose(ds_y_tot["volume"].values, volume, rtol=0.002)

        npt.assert_allclose(ds_y_tot["u_tot_volume"].values, u_tot_volume, rtol=0.1)

        gl = IdealGasLaw(
            prop,
            ["pressure", "temperature", "n_moles"],
            yvariable="volume",
            yunit="m^3",
            corr_dims=["x.y","time"],
        )
        ds_y_tot = gl.propagate_ds_total(ds)

        npt.assert_allclose(ds_y_tot["volume"].values, volume, rtol=0.002)

        npt.assert_allclose(ds_y_tot["u_tot_volume"].values, u_tot_volume, rtol=0.1)


    def test_hypernets_repeat_dim(self):
        prop = MCPropagation(3000, dtype="float32", parallel_cores=1, verbose=False)

        hmf = HypernetsMF(
            prop=prop,
            xvariables=[
                "digital_number",
                "gains",
                "dark_signal",
                "non_linearity_coefficients",
                "integration_time",
            ],
            yvariable="irradiance",
            yunit="W m^-2",
            corr_between=None,
            param_fixed=None,
        )

        hmf.setup(0.1)
        y = hmf.run(calib_data, L0data.variables,L0data)
        u_y_rand = hmf.propagate_random(L0data, calib_data)
        # print(u_y_rand,L0data)
        mask = np.where(
            (
                (L1data["wavelength"].values < 1350)
                | (L1data["wavelength"].values > 1450)
            )
        )

        u_y_syst_indep = hmf.propagate_specific("systematic_indep", L0data, calib_data)
        u_y_syst_corr = hmf.propagate_specific(
            "u_rel_systematic_corr_rad_irr", L0data, calib_data
        )

        u_y_syst = (u_y_syst_indep ** 2 + u_y_syst_corr ** 2) ** 0.5
        u_y_tot = (u_y_syst_indep ** 2 + u_y_syst_corr ** 2 + u_y_rand ** 2) ** 0.5

        ds_tot = hmf.propagate_ds_total(L0data, calib_data, store_unc_percent=True)
        mask = np.where(
            (np.isfinite(u_y_tot / y) & np.isfinite(ds_tot["u_rel_tot_irradiance"]))
        )[0]
        npt.assert_allclose(
            ds_tot["u_rel_tot_irradiance"][mask],
            u_y_tot[mask] / y[mask] * 100,
            rtol=0.05,
            atol=0.05,
        )

    def test_hypernets_repeat_dim2(self):
        prop = MCPropagation(3000, dtype="float32", parallel_cores=0, verbose=False)

        hmf = HypernetsMF(
            prop=prop,
            yvariable="irradiance",
            yunit="W m^-2",
            corr_between=None,
            repeat_dims="scan",
            corr_dims=-99,
        )
        hmf.setup(0.1)
        y = hmf.run(calib_data, L0data)
        u_y_rand = hmf.propagate_random(L0data, calib_data)
        # print(list(L1data.variables))
        mask = np.where(
            (
                (L1data["wavelength"].values < 1350)
                | (L1data["wavelength"].values > 1450)
            )
        )

        npt.assert_allclose(L1data["irradiance"].values, y, rtol=0.03)

        npt.assert_allclose(
            L1data["u_rel_random_irradiance"].values[mask][
                np.where(np.isfinite(u_y_rand[mask]))
            ],
            (u_y_rand[mask] / y[mask] * 100)[np.where(np.isfinite(u_y_rand[mask]))],
            rtol=0.03,
            atol=0.2,
        )

        ds_all = hmf.propagate_ds_all(L0data, calib_data, store_unc_percent=True)
        ds_main = hmf.propagate_ds(L0data, calib_data, store_unc_percent=True)
        ds_spec = hmf.propagate_ds_specific(
            ["random", "systematic_indep", "systematic_corr_rad_irr"],
            L0data,
            calib_data,
            store_unc_percent=True,
        )

        ds_main.to_netcdf("propagate_ds_example.nc")

        u_y_syst_indep = hmf.propagate_specific(
            "u_rel_systematic_indep", L0data, calib_data
        )
        u_y_syst_corr = hmf.propagate_specific(
            "u_rel_systematic_corr_rad_irr", L0data, calib_data, return_corr=False
        )

        u_y_syst = (u_y_syst_indep ** 2 + u_y_syst_corr ** 2) ** 0.5
        u_y_tot = (u_y_syst_indep ** 2 + u_y_syst_corr ** 2 + u_y_rand ** 2) ** 0.5

        npt.assert_allclose(
            L1data["u_rel_systematic_indep_irradiance"].values[mask],
            u_y_syst_indep[mask] / y[mask] * 100,
            rtol=0.03,
            atol=0.2,
        )

        npt.assert_allclose(
            L1data["u_rel_systematic_corr_rad_irr_irradiance"].values[mask],
            u_y_syst_corr[mask] / y[mask] * 100,
            rtol=0.03,
            atol=0.3,
        )

        npt.assert_allclose(
            L1data["u_rel_systematic_indep_irradiance"].values[mask],
            ds_spec["u_rel_systematic_indep_irradiance"].values[mask],
            rtol=0.03,
            atol=0.2,
        )

        npt.assert_allclose(
            L1data["u_rel_systematic_corr_rad_irr_irradiance"].values[mask],
            ds_spec["u_rel_systematic_corr_rad_irr_irradiance"].values[mask],
            rtol=0.03,
            atol=0.3,
        )

        npt.assert_allclose(
            L1data["u_rel_systematic_indep_irradiance"].values[mask],
            ds_all["u_rel_systematic_indep_irradiance"].values[mask],
            rtol=0.03,
            atol=0.2,
        )

        npt.assert_allclose(
            L1data["u_rel_systematic_corr_rad_irr_irradiance"].values[mask],
            ds_all["u_rel_systematic_corr_rad_irr_irradiance"].values[mask],
            rtol=0.03,
            atol=0.3,
        )
        # plt.plot(
        #     L1data["wavelength"][mask],
        #     ds_main["u_rel_str_irradiance"][mask] - (u_y_syst[mask] / y[mask]),
        #     "r-",
        # )
        # plt.show()

        npt.assert_allclose(
            ds_main["u_rel_ran_irradiance"][mask],
            u_y_rand[mask] / y[mask] * 100,
            rtol=0.03,
            atol=0.2,
        )
        npt.assert_allclose(
            ds_main["u_rel_str_irradiance"][mask],
            u_y_syst[mask] / y[mask] * 100,
            rtol=0.03,
            atol=0.2,
        )


if __name__ == "__main__":
    unittest.main()



"""Use Monte Carlo to propagate uncertainties"""

import numpy as np

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


def select_repeated_x(x, u_x, param_fixed, i, repeat_dims, repeat_shape):
    """
    Select one (index i) of multiple repeated entries and return the input quantities and uncertainties for that entry.

    :param x: list of input quantities (usually numpy arrays)
    :type x: list[array]
    :param u_x: list of uncertainties/covariances on input quantities (usually numpy arrays)
    :type u_x: list[array]
    :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
    :type param_fixed: list of bools, optional
    :param i: index of the repeated measurement
    :type i: int
    :param repeat_dims: dimension along which the measurements are repeated
    :type repeat_dims: int
    :param repeat_shape: shape of measurements along which to select repeats
    :type repeat_shape: tuple
    :return: list of input quantities, list of uncertainties for single measurement
    :rtype: list[array]. list[array]
    """
    xb = np.zeros(len(x), dtype=object)
    u_xb = np.zeros(len(u_x), dtype=object)
    for j in range(len(x)):
        selected = False
        if param_fixed is not None:
            if param_fixed[j] == True:
                xb[j] = x[j]
                u_xb[j] = u_x[j]
                selected = True
        if not selected:
            index = list(np.ndindex(repeat_shape))[i]
            xb[j] = x[j]
            u_xb[j] = u_x[j]
            for idim in range(len(repeat_dims)):
                repeat_axis = repeat_dims[idim]
                ii = index[idim]
                if len(xb[j].shape) > repeat_axis:
                    sli = tuple(
                        [
                            ii if (idim == repeat_axis) else slice(None)
                            for idim in range(xb[j].ndim)
                        ]
                    )
                    xb[j] = xb[j][sli]
                    u_xb[j] = u_xb[j][sli]
                elif len(xb[j]) > 1:
                    try:
                        xb[j] = xb[j][ii]
                        u_xb[j] = u_xb[j][ii]
                    except:
                        pass

    return xb, u_xb



import numpy as np

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


def bell_relative(len, n, sigma=None):
    if sigma is None:
        sigma = (n / 2 - 1) / 3 ** 0.5
    corr = np.eye(len)
    for i in range(n):
        idx_row = np.arange(i, len)
        idx_col = np.arange(len - i)
        corr[idx_row, idx_col] = np.exp(-0.5 * (i / sigma) ** 2)
        corr[idx_col, idx_row] = np.exp(-0.5 * (i / sigma) ** 2)
    return corr


def triangular_relative(len, n):
    corr = np.eye(len)
    for i in range(n):
        idx_row = np.arange(i, len)
        idx_col = np.arange(len - i)
        corr[idx_row, idx_col] = (n - i) / n
        corr[idx_col, idx_row] = (n - i) / n
    return corr



"""Class to make templates for digital effects tables for measurand"""

import warnings
from abc import ABC

import numpy as np
import xarray as xr

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


class DigitalEffectsTableTemplates(ABC):
    def __init__(self, yvariable, yunit, output_vars):
        """
        Initialise DigitalEffectsTableTemplates

        :param yvariable: name of the measurand variable
        :type yvariable: string
        :param yunit: unit of the measurand
        :type yunit: string
        """
        if isinstance(yvariable,str):
            self.yvariable = [yvariable]
            self.yunit = [yunit]
        else:
            self.yvariable = yvariable
            self.yunit = yunit

        self.output_vars = output_vars

    def make_template_main(
        self,
        dims,
        dim_sizes,
            str_corr_dims=[],
            str_repeat_noncorr_dims=[],
        store_unc_percent=False,
        repeat_dim_err_corrs=[],
    ):
        """
        Make the digital effects table template for the case where random, systematic and structured uncertainties are propagated seperately

        :param dims: list of dimensions
        :type dims: list
        :param u_xvar_ref: reference uncertainty component that is used to populate repeated dims
        :type u_xvar_ref: xarray.Variable
        :param str_repeat_noncorr_dims: set to (list of) string(s) with dimension name(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to [], for which there is no reduction in dimensionality..
        :type str_repeat_noncorr_dims: str or list of str, optional
        :return: measurand digital effects table template to be used by obsarray
        :rtype: dict
        """
        template = {}

        for i in range(self.output_vars):
            err_corr, custom_err_corr = self.set_errcorr_shape(
                dims[i],
                dim_sizes,
                "err_corr_str_" + self.yvariable[i],
                str_repeat_noncorr_dims=str_repeat_noncorr_dims,
                str_corr_dims=str_corr_dims,
                repeat_dim_err_corr=repeat_dim_err_corrs,
            )

            if store_unc_percent:
                units = "%"
            else:
                units = self.yunit[i]

            template[self.yvariable[i]] = {
                    "dtype": np.float32,
                    "dim": dims[i],
                    "attributes": {
                        "units": self.yunit[i],
                        "unc_comps": [
                            self.make_ucomp_name(self.yvariable[i],
                                "ran", store_unc_percent=store_unc_percent
                            ),
                            self.make_ucomp_name(self.yvariable[i],
                                "sys", store_unc_percent=store_unc_percent
                            ),
                            self.make_ucomp_name(self.yvariable[i],
                                "str", store_unc_percent=store_unc_percent
                            ),
                        ],
                    },
                }
            template[self.make_ucomp_name(self.yvariable[i],"ran", store_unc_percent=store_unc_percent)] = {
                    "dtype": np.float32,
                    "dim": dims[i],
                    "attributes": {
                        "units": units,
                        "err_corr": [
                            {"dim": dim, "form": "random", "params": [], "units": []}
                            for dim in dims
                        ],
                    },
                }
            template[self.make_ucomp_name(self.yvariable[i],"sys", store_unc_percent=store_unc_percent)] ={
                    "dtype": np.float32,
                    "dim": dims[i],
                    "attributes": {
                        "units": units,
                        "err_corr": [
                            {"dim": dim, "form": "systematic", "params": [], "units": []}
                            for dim in dims
                        ],
                    },
                }
            template[self.make_ucomp_name(self.yvariable[i],"str", store_unc_percent=store_unc_percent)] = {
                    "dtype": np.float32,
                    "dim": dims[i],
                    "attributes": {"units": units, "err_corr": err_corr},
                }


            if custom_err_corr is not None:
                for key in custom_err_corr.keys():
                    if "err_corr_str_" + self.yvariable[i] in key:
                        template[key] = custom_err_corr[key]

        return template

    def make_template_tot(
        self,
        dims,
        dim_sizes,
        str_corr_dims=[],
        str_repeat_noncorr_dims=[],
        store_unc_percent=False,
        repeat_dim_err_corrs=[],
    ):
        """
        Make the digital effects table template for the case where uncertainties are combined and only the total uncertainty is returned.

        :param dims: list of dimensions
        :type dims: list
        :param u_xvar_ref: reference uncertainty component that is used to populate repeated dims
        :type u_xvar_ref: xarray.Variable
        :return: measurand digital effects table template to be used by obsarray
        :rtype: dict
        """
        template = {}

        for i in range(self.output_vars):
            err_corr, custom_err_corr = self.set_errcorr_shape(
                dims[i],
                dim_sizes,
                "err_corr_tot_" + self.yvariable[i],
                str_repeat_noncorr_dims=str_repeat_noncorr_dims,
                str_corr_dims=str_corr_dims,
                repeat_dim_err_corr=repeat_dim_err_corrs,
            )

            if store_unc_percent:
                units = "%"
            else:
                units = self.yunit[i]

            template[self.yvariable[i]]={
                    "dtype": np.float32,
                    "dim": dims[i],
                    "attributes": {
                        "units": self.yunit[i],
                        "unc_comps": [
                            self.make_ucomp_name(self.yvariable[i],"tot", store_unc_percent=store_unc_percent)
                        ],
                    },
                }
            template[self.make_ucomp_name(self.yvariable[i],"tot", store_unc_percent=store_unc_percent)]={
                    "dtype": np.float32,
                    "dim": dims[i],
                    "attributes": {"units": units, "err_corr": err_corr},
                }

            if custom_err_corr is not None:
                for key in custom_err_corr.keys():
                    if "err_corr_tot_" + self.yvariable[i] in key:
                        template[key] = custom_err_corr[key]

        # if self.output_vars>1:
        #     template["corr_between_vars"]={
        #         "dtype": np.float32,
        #         "dim": dims,
        #         "attributes": {"units": units, "err_corr": err_corr},
        #     }

        return template

    def make_template_specific(
        self,
        comp_list,
        dims,
        dim_sizes,
            str_corr_dims=[],
            str_repeat_noncorr_dims=[],
        store_unc_percent=False,
        repeat_dim_err_corrs=[],
        simple_random=True,
        simple_systematic=True,
    ):
        """
        Make the digital effects table template for the case where uncertainties are combined and only the total uncertainty is returned.

        :param dims: list of dimensions
        :type dims: list
        :param u_xvar_ref: reference uncertainty component that is used to populate repeated dims
        :type u_xvar_ref: xarray.Variable
        :return: measurand digital effects table template to be used by obsarray
        :rtype: dict
        """
        template = {}

        for i in range(self.output_vars):
            if store_unc_percent:
                units = "%"
            else:
                units = self.yunit[i]

            template[self.yvariable[i]] = {
                    "dtype": np.float32,
                    "dim": dims[i],
                    "attributes": {
                        "units": self.yunit[i],
                        "unc_comps": [
                            self.make_ucomp_name(self.yvariable[i],comp, store_unc_percent=store_unc_percent)
                            for comp in comp_list
                        ],
                    },
                }

            for ic, comp in enumerate(comp_list):
                if comp == "random" and simple_random:
                    template[
                        self.make_ucomp_name(self.yvariable[i],comp, store_unc_percent=store_unc_percent)
                    ] = {
                        "dtype": np.float32,
                        "dim": dims[i],
                        "attributes": {
                            "units": units,
                            "err_corr": [
                                {"dim": dim, "form": "random", "params": [], "units": []}
                                for dim in dims[i]
                            ],
                        },
                    }

                elif comp == "systematic" and simple_systematic:
                    template[
                        self.make_ucomp_name(self.yvariable[i], comp, store_unc_percent=store_unc_percent)
                    ] = {
                        "dtype": np.float32,
                        "dim": dims[i],
                        "attributes": {
                            "units": units,
                            "err_corr": [
                                {
                                    "dim": dim,
                                    "form": "systematic",
                                    "params": [],
                                    "units": [],
                                }
                                for dim in dims[i]
                            ],
                        },
                    }

                else:
                    err_corr, custom_err_corr = self.set_errcorr_shape(
                        dims[i],
                        dim_sizes,
                        "err_corr_" + comp + "_" + self.yvariable[i],
                        str_repeat_noncorr_dims=str_repeat_noncorr_dims,
                        str_corr_dims=str_corr_dims,
                        repeat_dim_err_corr=repeat_dim_err_corrs[ic],
                    )

                    template[
                        self.make_ucomp_name(self.yvariable[i], comp, store_unc_percent=store_unc_percent)
                    ] = {
                        "dtype": np.float32,
                        "dim": dims[i],
                        "attributes": {"units": units, "err_corr": err_corr},
                    }
                    if custom_err_corr is not None:
                        for key in custom_err_corr.keys():
                            if "err_corr_" + comp + "_" + self.yvariable[i] in key:
                                template[key] = custom_err_corr[key]

        return template

    def make_ucomp_name(self, var, ucomp, store_unc_percent=False):
        if store_unc_percent:
            uvarname_start = "u_rel_"
        else:
            uvarname_start = "u_"
        return uvarname_start + ucomp + "_" + var

    def set_errcorr_shape(
        self,
        dims,
        dim_sizes,
        err_corr_name,
        str_repeat_noncorr_dims=[],
        str_corr_dims=[],
        repeat_dim_err_corr=[],
    ):
        """
        Function to work out which di
        :param dims: list of dimensions
        :type dims: list
        :param u_xvar_ref: reference uncertainty component that is used to populate repeated dims
        :type u_xvar_ref: xarray.Variable
        :param err_corr_name:
        :type err_corr_name:
        :param str_repeat_noncorr_dims: set to (list of) string(s) with dimension name(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to [], for which there is no reduction in dimensionality..
        :type str_repeat_noncorr_dims: str or list of str, optional
        :return: measurand digital effects table template to be used by obsarray
        :rtype: dict
        """
        err_corr_list = []
        custom_corr_dims = []
        custom_err_corr_dict = None

        # loop through all dimensions, and copy the ones that are repeat dims
        for dim in dims:
            if dim in str_repeat_noncorr_dims:
                err_corr_list.append(repeat_dim_err_corr[dim])
            else:
                custom_corr_dims.append(dim)

        if len(str_corr_dims)==0:
            if len(custom_corr_dims) > 0:
                if len(custom_corr_dims) == 1:
                    corrdim = custom_corr_dims[0]
                    custom_corr_dims = custom_corr_dims[0]
                else:
                    corrdim = ".".join(custom_corr_dims)
                    # dim_sizes[corrdim] = 1
                    # for cust_dim in custom_corr_dims:
                    #     dim_sizes[corrdim] *= dim_sizes[cust_dim]

                custom_err_corr_dict = {err_corr_name:{
                    "dtype": np.float32,
                    "dim": [corrdim, corrdim],
                    "attributes": {"units": ""},
                }}

            # make a combined custom form for the variables that are not repeated dims
            if len(custom_corr_dims) > 0:
                err_corr_list.append(
                    {
                        "dim": custom_corr_dims,
                        "form": "err_corr_matrix",
                        "params": [err_corr_name],
                        "units": [],
                    }
                )
        else:
            custom_err_corr_dict={}
            for corrdim in str_corr_dims:
                custom_err_corr_dict[err_corr_name+"_"+corrdim]={
                    "dtype": np.float32,
                    "dim": [corrdim, corrdim],
                    "attributes": {"units": ""},
                }

                err_corr_list.append(
                    {
                        "dim": corrdim.split("."),
                        "form": "err_corr_matrix",
                        "params": [err_corr_name+"_"+corrdim],
                        "units": [],
                    }
                )

        return err_corr_list, custom_err_corr_dict

    def remove_unc_component(self, ds, variable, u_comp, err_corr_comp=None):
        """
        Function to remove an uncertainty component from a dataset

        :param ds: dataset from which uncertainty should be removed
        :type ds: xr.Dataset
        :param variable: variable from which uncertainty should be removed
        :type variable: str
        :param u_comp: name of uncertainty component that should be removed
        :type u_comp: str
        :param err_corr_comp: optionally, the name of the error correlation matrix component associated with this uncertainty component can be provided, so it is removed as well.
        :type err_corr_comp: str
        :return: dataset which uncertainty component removed
        :rtype: xr.Dataset
        """
        ds[variable].attrs["unc_comps"].remove(u_comp)
        try:
            ds = ds.drop(u_comp)
            if err_corr_comp is not None:
                ds = ds.drop(err_corr_comp)
        except:
            pass
        return ds

    def join_with_preexisting_ds(self, ds, ds_pre, drop=None):
        """
        Function to combine digital effects table, with previously populated dataset.
        Only the measurand is overwritten.

        :param ds: digital effects table for measurand, created by punpy
        :type ds: xarray.Dataset
        :param ds_pre: previously populated dataset, to be combined with digital effects table
        :type ds_pre: xarray.Dataset
        :param drop: list of variables to drop
        :param drop: List(str)
        :return: merged digital effects table
        :rtype: xarray.Dataset
        """
        for var in ds.variables:
            ds[var].encoding = ds_pre[var].encoding
            err_corr_warn = False
            for key in ds_pre[var].attrs.keys():
                if "err_corr" in key:
                    try:
                        if ds[var].attrs[key] == ds_pre[var].attrs[key]:
                            err_corr_warn = True
                    except:
                        err_corr_warn = True

                else:
                    ds[var].attrs[key] = ds_pre[var].attrs[key]
            if err_corr_warn:
                warnings.warn(
                    "The returned dataset has a different error correlation parameterisation for variable %s than the provided ds_out_pre (possibly just the order changed)."
                    % var
                )

        if drop is not None:
            if drop in ds_pre.variables:
                ds_pre = ds_pre.drop(drop)
        for var in ds.variables:
            try:
                ds_pre = ds_pre.drop(var)
            except:
                continue

        ds_out = xr.merge([ds, ds_pre])
        ds_out.attrs = ds_pre.attrs

        return ds_out



"""
Tests for mc propagation class
"""

import unittest

import comet_maths as cm
import numpy as np
import numpy.testing as npt

from punpy.mc.mc_propagation import MCPropagation

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "14/4/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


def function(x1, x2):
    return x1 ** 2 - 10 * x2 + 30.0


x1 = np.ones(200) * 10
x2 = np.ones(200) * 30
x1err = np.ones(200)
x2err = 2 * np.ones(200)

xs = [x1, x2]
xerrs = [x1err, x2err]

# below, the higher order Taylor expansion terms have been taken into account, and amount to 2.
yerr_uncorr = 802 ** 0.5 * np.ones(200)
yerr_corr = 2 ** 0.5 * np.ones(200)


def functionb(x1, x2):
    return 2 * x1 - x2

def functionb_fail(x1, x2):
    zero_or_one=np.random.choice([0,1],p=[0.1,0.9])
    with np.errstate(divide='raise'):
        try:
            return 2 * x1 - x2/zero_or_one
        except:
            return np.nan

def functionb_failnan(x1, x2):
    zero_or_one=np.random.choice([0,1],p=[0.1,0.9])
    return 2 * x1 - x2/zero_or_one


x1b = np.ones((20, 30)) * 50
x2b = np.ones((20, 30)) * 30
x1errb = np.ones((20, 30))
x2errb = 2 * np.ones((20, 30))

xsb = np.array([x1b, x2b])
xerrsb = np.array([x1errb, x2errb])

yerr_uncorrb = 8 ** 0.5 * np.ones((20, 30))
yerr_corrb = np.zeros((20, 30))


def functionc(x1, x2, x3):
    return x1 + 4 * x2 - 2 * x3


x1c = np.ones(200) * 10
x2c = np.ones(200) * 10
x3c = np.ones(200) * 10

x1errc = 12 * np.ones(200)
x2errc = 5 * np.ones(200)
x3errc = np.zeros(200)

xsc = np.array([x1c, x2c, x3c])
xerrsc = np.array([x1errc, x2errc, x3errc])
corr_c = np.array([[1, 0.9999999, 0], [0.99999999, 1.0, 0], [0.0, 0.0, 1.0]])
yerr_uncorrc = 544 ** 0.5 * np.ones(200)
yerr_corrc = 1024 ** 0.5 * np.ones(200)


def functiond(x1, x2):
    return 2 * x1 - x2, 2 * x1 + x2


x1d = np.ones((20, 3, 4)) * 50
x2d = np.ones((20, 3, 4)) * 30
x1errd = np.ones((20, 3, 4))
x2errd = 2 * np.ones((20, 3, 4))

xsd = [x1d, x2d]
xerrsd = [x1errd, x2errd]
corr_d = np.ones(
    (2, 2)
)  # np.array([[1,0.9999999,0.9999999],[0.99999999,1.,0.99999999],[0.9999999,0.9999999,1.]])

yerr_uncorrd = [
    np.array(8 ** 0.5 * np.ones((20, 3, 4))),
    np.array(8 ** 0.5 * np.ones((20, 3, 4))),
]
yerr_corrd = [np.zeros((20, 3, 4)), 16 ** 0.5 * np.ones((20, 3, 4))]


def functione(x1, x2):
    return 2 * x1 - x2, np.mean(2 * x1 + x2, axis=0)


class TestMCPropagation(unittest.TestCase):
    """
    Class for unit tests
    """

    def test_propagate_random_1D(self):
        prop = MCPropagation(40000, parallel_cores=0)

        uf, ucorr = prop.propagate_random(function, xs, xerrs, return_corr=True)
        npt.assert_allclose(ucorr, np.eye(len(ucorr)), atol=0.06)
        npt.assert_allclose(uf, yerr_uncorr, rtol=0.06)
        ucov = cm.convert_corr_to_cov(ucorr, uf)
        ucorr2 = cm.convert_cov_to_corr(ucov, uf)
        npt.assert_allclose(ucorr, ucorr2, atol=0.01)

        uf = prop.propagate_random(function, xs, xerrs, corr_between=np.ones((2, 2)))
        npt.assert_allclose(uf, yerr_corr, atol=0.1)

        uf, ucorr, yvalues, xvalues = prop.propagate_random(
            function, xs, xerrs, return_corr=True, return_samples=True
        )

        npt.assert_allclose(uf, yerr_uncorr, rtol=0.06)

        uf, yvalues, xvalues = prop.propagate_random(
            function, xs, xerrs, corr_between=np.ones((2, 2)), return_samples=True
        )
        npt.assert_allclose(uf, yerr_corr, atol=0.1)

        uf, yvalues, xvalues = prop.propagate_standard(
            function,
            xs,
            xerrs,
            corr_x=["rand", "rand"],
            corr_between=np.ones((2, 2)),
            return_samples=True,
        )
        npt.assert_allclose(uf, yerr_corr, atol=0.1)

        uf, yvalues, xvalues = prop.propagate_random(
            function,
            xs,
            xerrs,
            corr_x=["rand", None],
            corr_between=np.ones((2, 2)),
            return_samples=True,
        )
        npt.assert_allclose(uf, yerr_corr, atol=0.1)

        uf, yvalues, xvalues = prop.propagate_random(
            function,
            xs,
            xerrs,
            corr_x=[np.diag(np.ones(xs[0].shape)), None],
            corr_between=np.ones((2, 2)),
            return_samples=True,
        )
        npt.assert_allclose(uf, yerr_corr, atol=0.1)

    def test_propagate_random_2D(self):
        prop = MCPropagation(20000, parallel_cores=4)
        ufb, ucorrb = prop.propagate_random(
            functionb, xsb, xerrsb, return_corr=True, repeat_dims=1
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        prop = MCPropagation(20000, parallel_cores=2, dtype=np.float32)

        ufb, ucorrb = prop.propagate_random(
            functionb, xsb, xerrsb, corr_x=[None, np.eye(len(xsb[0]))], return_corr=True
        )

        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        ufb, ucorrb = prop.propagate_random(
            functionb,
            xsb,
            xerrsb,
            corr_x=[None, np.eye(len(xsb[0] * xsb[1]))],
            return_corr=True,
        )

        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        ufb, ucorrb = prop.propagate_random(
            functionb, xsb, xerrsb, return_corr=True, corr_dims=1
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        ufb = prop.propagate_random(
            functionb, xsb, xerrsb, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.06)

        ufb, yvaluesb, xvaluesb = prop.propagate_random(
            functionb,
            xsb,
            xerrsb,
            corr_between=np.ones((2, 2)),
            return_samples=True,
            repeat_dims=1,
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.07)

    def test_propagate_random_1D_3var(self):
        prop = MCPropagation(20000, parallel_cores=1)
        ufc, ucorrc = prop.propagate_random(functionc, xsc, xerrsc, return_corr=True)
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.06)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.06)

        ufc = prop.propagate_random(functionc, xsc, xerrsc, corr_between=corr_c)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.06)

        ufc, yvaluesc, xvaluesc = prop.propagate_random(
            functionc, xsc, xerrsc, corr_between=corr_c, return_samples=True
        )
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.06)

    def test_propagate_random_3D_2out(self):
        prop = MCPropagation(20000, parallel_cores=0, verbose=True)
        ufd, ucorrd, corr_out = prop.propagate_random(
            functiond, xsd, xerrsd, return_corr=True, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], rtol=0.06)

        ufd, ucorrd, corr_out, yvalues, xvalues = prop.propagate_random(
            functiond,
            xsd,
            xerrsd,
            return_corr=True,
            output_vars=2,
            return_samples=True,
            repeat_dims=[2, 1],
        )
        npt.assert_allclose(ucorrd[1], np.eye(len(ucorrd[1])), atol=0.06)

        ufd = prop.propagate_random(
            functiond, xsd, xerrsd, corr_between=corr_d, output_vars=2
        )
        npt.assert_allclose(ufd[0], yerr_corrd[0], atol=0.06)
        npt.assert_allclose(ufd[1], yerr_corrd[1], rtol=0.06)

        ufd, yvaluesd, xvaluesd = prop.propagate_random(
            functiond,
            xsd,
            xerrsd,
            corr_between=corr_d,
            return_samples=True,
            output_vars=2,
        )
        npt.assert_allclose(ufd[0], yerr_corrd[0], atol=0.1)
        npt.assert_allclose(ufd[1], yerr_corrd[1], atol=0.1)

        ufd = prop.propagate_random(
            functiond,
            xsd,
            xerrsd,
            corr_between=corr_d,
            return_samples=False,
            output_vars=2,
            repeat_dims=0,
        )
        npt.assert_allclose(ufd[0], yerr_corrd[0], atol=0.1)
        npt.assert_allclose(ufd[1], yerr_corrd[1], atol=0.1)

    def test_propagate_systematic_2D_corrdict(self):
        prop = MCPropagation(20000)

        corrb = [{"0":np.eye(len(xerrb[:,0].ravel())),"1":np.eye(len(xerrb[0].ravel()))} for xerrb in xerrsb]
        ufb, ucorrb = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=corrb,
            return_corr=True,
        )

        ufb2, ucorrb2 = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=[{"0":np.eye(len(xerrsb[0])),"1":"syst"}, "syst"],
            return_corr=True,
        )

        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)
        npt.assert_allclose(ufb2, yerr_uncorrb, rtol=0.06)

    def test_propagate_systematic_1D(self):
        prop = MCPropagation(30000)

        corr = (np.ones_like(x1err) + np.eye(len(x1err))) / 2
        uf, ucorr = prop.propagate_systematic(
            function,
            xs,
            [x1err, None],
            corr_x=[corr, None],
            return_corr=True,
            fixed_corr_var=True,
        )
        npt.assert_allclose(ucorr, corr, atol=0.0001)

        uf, ucorr = prop.propagate_systematic(function, xs, xerrs, return_corr=True)
        npt.assert_allclose(ucorr, np.ones_like(ucorr), atol=0.06)
        npt.assert_allclose(uf, yerr_uncorr, rtol=0.06)

        uf = prop.propagate_systematic(
            function, xs, xerrs, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(uf, yerr_corr, rtol=0.06)

    def test_propagate_systematic_2D(self):
        prop = MCPropagation(30000)
        ufb, ucorrb = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        ufb, ucorrb = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True, repeat_dims=1
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.06)

        ufb, ucorrb = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True, corr_dims=0
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.06)

        ufb = prop.propagate_systematic(
            functionb, xsb, xerrsb, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

    def test_propagate_systematic_1D_3var(self):
        prop = MCPropagation(30000)
        ufc, ucorrc = prop.propagate_systematic(
            functionc, xsc, xerrsc, return_corr=True
        )
        npt.assert_allclose(ucorrc, np.ones_like(ucorrc), atol=0.06)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.06)

        ufc = prop.propagate_systematic(functionc, xsc, xerrsc, corr_between=corr_c)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.06)

        xsc2 = np.array([x1c, 10.0, x3c])
        xerrsc2 = np.array([x1errc, 5.0, x3errc])

        ufc, ucorrc = prop.propagate_systematic(
            functionc, xsc2, xerrsc2, return_corr=True, param_fixed=[False, True, False]
        )
        npt.assert_allclose(ucorrc, np.ones_like(ucorrc), atol=0.06)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.06)

        xsc2 = np.array([x1c, 10.0, x3c])
        xerrsc2 = np.array([x1errc, 5.0, x3errc])

        ufc = prop.propagate_systematic(functionc, xsc2, xerrsc2)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.06)

    def test_propagate_systematic_3D_2out(self):
        prop = MCPropagation(30000, parallel_cores=2)
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.06)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_dims=0, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.06)

        prop = MCPropagation(30000)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_dims=1, output_vars=2
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.06)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_dims=2, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.06)

    def test_propagate_cov_1D(self):
        prop = MCPropagation(25000)

        cov = [
            cm.convert_corr_to_cov(np.eye(len(xerr.ravel())), xerr) for xerr in xerrs
        ]
        uf, ucorr = prop.propagate_cov(function, xs, cov, return_corr=True)
        npt.assert_allclose(ucorr, np.eye(len(ucorr)), atol=0.06)
        npt.assert_allclose(uf, yerr_uncorr, rtol=0.06)

        cov = [
            cm.convert_corr_to_cov(
                np.ones((len(xerr.ravel()), len(xerr.ravel()))) + np.eye(len(xerr)),
                xerr,
            )
            for xerr in xerrs
        ]
        uf, ucorr = prop.propagate_cov(function, xs, cov, return_corr=True)
        npt.assert_allclose(uf, yerr_uncorr * 2 ** 0.5, rtol=0.06)

        cov = [
            cm.convert_corr_to_cov(np.eye(len(xerr.ravel())), xerr) for xerr in xerrs
        ]
        uf, ucorr = prop.propagate_cov(
            function, xs, cov, return_corr=True, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ucorr, np.eye(len(ucorr)), atol=0.06)
        npt.assert_allclose(uf, yerr_corr, atol=0.15)

    def test_propagate_cov_2D(self):
        prop = MCPropagation(20000)

        covb = [
            cm.convert_corr_to_cov(np.eye(len(xerrb.ravel())), xerrb)
            for xerrb in xerrsb
        ]

        ufb, ucorrb = prop.propagate_cov(functionb, xsb, covb, return_corr=True)
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        # covb = [
        #     cm.convert_corr_to_cov(np.eye(len(xerrb[:, 0].ravel())), xerrb[:, 0])
        #     for xerrb in xerrsb
        # ]
        #
        # ufb, ucorrb = prop.propagate_cov(
        #     functionb, xsb, covb, return_corr=True, repeat_dims=1
        # )
        # npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        # npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        xsb2 = [50, xsb[1]]
        covb = [
            1,
            cm.convert_corr_to_cov(
                np.ones((len(xerrsb[1].ravel()), len(xerrsb[1].ravel()))),
                xerrsb[1],
            ),
        ]

        ufb, ucorrb = prop.propagate_cov(
            functionb,
            xsb2,
            covb,
            return_corr=True,
            param_fixed=[True, False],
        )
        npt.assert_allclose(ucorrb, np.ones((len(ucorrb), len(ucorrb))), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        covb = [
            cm.convert_corr_to_cov(
                np.ones((len(xerrb.ravel()), len(xerrb.ravel())))
                + np.eye(len(xerrb.ravel())),
                xerrb,
            )
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_cov(functionb, xsb, covb, return_corr=True)
        npt.assert_allclose(ufb, yerr_uncorrb * 2 ** 0.5, rtol=0.06)

        covb = [
            cm.convert_corr_to_cov(np.eye(len(xerrb.ravel())), xerrb)
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_cov(
            functionb, xsb, covb, return_corr=True, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_corrb, atol=0.1)

    def test_propagate_cov_1D_3var(self):
        prop = MCPropagation(20000)
        covc = [
            cm.convert_corr_to_cov(np.eye(len(xerrc.ravel())), xerrc)
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(functionc, xsc, covc, return_corr=True)
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.06)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.06)

        covc = [
            cm.convert_corr_to_cov(
                np.ones((len(xerrc.ravel()), len(xerrc.ravel()))) + np.eye(len(xerrc)),
                xerrc,
            )
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(functionc, xsc, covc, return_corr=True)
        npt.assert_allclose(ufc, yerr_uncorrc * 2 ** 0.5, rtol=0.06)

        covc = [
            cm.convert_corr_to_cov(np.eye(len(xerrc.ravel())), xerrc)
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(
            functionc, xsc, covc, return_corr=True, corr_between=corr_c
        )
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.06)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.06)

    def test_propagate_cov_3D_2out(self):
        prop = MCPropagation(20000)

        covd = [
            cm.convert_corr_to_cov(np.eye(len(xerrd.ravel())), xerrd)
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_cov(
            functiond, xsd, covd, return_corr=True, corr_dims=0, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], rtol=0.06)

        covd = [
            cm.convert_corr_to_cov(
                np.ones((len(xerrd.ravel()), len(xerrd.ravel()))), xerrd
            )
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_cov(
            functiond, xsd, covd, return_corr=True, corr_dims=0, output_vars=2
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], rtol=0.06)

    def test_propagate_syst_corr_2D(self):
        prop = MCPropagation(20000)

        corrb = [np.eye(len(xerrb[0].ravel())) for xerrb in xerrsb]
        ufb, ucorrb = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=corrb,
            repeat_dims=0,
            corr_dims=1,
            return_corr=True,
        )
        ufb, ucorrb2 = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=["rand", "rand"],
            repeat_dims=0,
            corr_dims=1,
            return_corr=True,
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ucorrb2, np.eye(len(ucorrb2)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        corrb = [
            np.ones((len(xerrb[0].ravel()), len(xerrb[0].ravel()))) for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=corrb,
            repeat_dims=0,
            corr_dims=1,
            return_corr=True,
        )
        ufb, ucorrb2 = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=["syst", None],
            repeat_dims=0,
            corr_dims=1,
            return_corr=True,
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.06)
        npt.assert_allclose(ucorrb2, np.ones_like(ucorrb2), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

    def test_propagate_syst_corr_3D_2out(self):
        prop = MCPropagation(20000, parallel_cores=1)

        corrd = [np.eye(len(xerrd[:, 0, 0].ravel())) for xerrd in xerrsd]
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond,
            xsd,
            xerrsd,
            corr_x=corrd,
            return_corr=True,
            corr_dims=0,
            repeat_dims=[1, 2],
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], rtol=0.06)

        prop = MCPropagation(20000, parallel_cores=1)

        corrd = [np.eye(len(xerrd[:, 0, 0].ravel())) for xerrd in xerrsd]
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond,
            xsd,
            xerrsd,
            corr_x=corrd,
            return_corr=True,
            corr_dims=0,
            repeat_dims=1,
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], rtol=0.06)

        xsd2 = [x1d, 1.0]
        xerrsd2 = [x1errd, 2.0]

        prop = MCPropagation(20000, parallel_cores=1)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functione,
            xsd2,
            xerrsd2,
            return_corr=True,
            corr_dims=0,
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)

        prop = MCPropagation(20000, parallel_cores=3)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functione,
            xsd2,
            xerrsd2,
            return_corr=True,
            corr_dims=0,
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functione,
            xsd2,
            xerrsd2,
            return_corr=True,
            corr_dims=0,
            output_vars=2,
            repeat_dims=1,
            param_fixed=[False, True],
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)

        prop = MCPropagation(20000, parallel_cores=1, dtype=np.float32)

        corrd = [
            np.ones((len(xerrd[:, 0, 0].ravel()), len(xerrd[:, 0, 0].ravel())))
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond,
            xsd,
            xerrsd,
            corr_x=corrd,
            return_corr=True,
            corr_dims=0,
            repeat_dims=[1, 2],
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        print(ufd[0][0, 0, 0].dtype)
        self.assertIsInstance(ufd[0][0, 0, 0], np.float32)

        ufd, ucorrd = prop.propagate_systematic(
            functionb,
            xsd,
            xerrsd,
            corr_x=corrd,
            return_corr=True,
            corr_dims=0,
            repeat_dims=[1, 2],
            output_vars=1,
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.01)

    def test_separate_processing(self):
        prop = MCPropagation(20000, parallel_cores=0)

        corrd = [np.eye(len(xerrd.ravel())) for xerrd in xerrsd]

        MC_x = prop.generate_MC_sample(xsd, xerrsd, corrd)
        MC_y1 = prop.run_samples(functiond, MC_x, output_vars=2, start=0, end=10000)
        MC_y2 = prop.run_samples(functiond, MC_x, output_vars=2, start=10000, end=20000)
        MC_y = prop.combine_samples([MC_y1, MC_y2])

        ufd, ucorrd, corr_out = prop.process_samples(
            MC_x, MC_y, return_corr=True, corr_dims=0, output_vars=2
        )

        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], rtol=0.06)

        prop = MCPropagation(20000, parallel_cores=0)

        corrd = [np.eye(len(xerrd[:, 0, 0].ravel())) for xerrd in xerrsd]

        MC_x = prop.generate_MC_sample(xsd, xerrsd, corr_x=["rand", "rand"])
        MC_y1 = prop.run_samples(functiond, MC_x, output_vars=2, start=0, end=10000)
        MC_y2 = prop.run_samples(functiond, MC_x, output_vars=2, start=10000, end=20000)
        MC_y = prop.combine_samples([MC_y1, MC_y2])

        ufd, ucorrd, corr_out = prop.process_samples(
            MC_x, MC_y, return_corr=True, corr_dims=0, output_vars=2
        )

        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], rtol=0.06)

        prop = MCPropagation(20000, parallel_cores=4)

        corrd = [np.eye(len(xerrd[:, 0, 0].ravel())) for xerrd in xerrsd]

        MC_x = prop.generate_MC_sample(xsd, xerrsd, corr_x=["rand", "rand"])
        MC_y1 = prop.run_samples(functiond, MC_x, output_vars=2, start=0, end=10000)
        MC_y2 = prop.run_samples(functiond, MC_x, output_vars=2, start=10000, end=20000)
        MC_y = prop.combine_samples([MC_y1, MC_y2])

        ufd, ucorrd, corr_out = prop.process_samples(
            MC_x, MC_y, return_corr=True, corr_dims=0, output_vars=2
        )

        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], rtol=0.06)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], rtol=0.06)

    def test_failed_processing(self):
        prop = MCPropagation(10000, parallel_cores=4,verbose=True)
        ufb, ucorrb = prop.propagate_random(
            functionb_fail, xsb, xerrsb, return_corr=True
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        prop = MCPropagation(10000, parallel_cores=4,verbose=True)
        ufb, ucorrb = prop.propagate_random(
            functionb_failnan, xsb, xerrsb, return_corr=True
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)


    def test_perform_checks(self):
        prop = MCPropagation(20000)
        corrc = [
            np.ones((len(xerrc[0].ravel()), len(xerrc[0].ravel()))) for xerrc in xerrsc
        ]
        corrd = [
            np.ones((len(xerrd[0].ravel()), len(xerrd[0].ravel()))) for xerrd in xerrsd
        ]

        out = prop.perform_checks(
            functionc,
            xsc,
            xerrsc,
            corr_x=corrc,
            repeat_dims=0,
            corr_dims=1,
            output_vars=1,
            fixed_corr_var=None,
            param_fixed=None,
            refyvar=0,
        )

        out = prop.perform_checks(
            functiond,
            xsd,
            xerrsd,
            corr_x=corrd,
            repeat_dims=[0, 1],
            corr_dims=2,
            output_vars=2,
            fixed_corr_var=None,
            param_fixed=[True, False, False],
            refyvar=0,
        )
        try:
            out = prop.perform_checks(
                functionc,
                xsc,
                xerrsc,
                corr_x=corrc,
                repeat_dims=0,
                corr_dims=0,
                output_vars=1,
                fixed_corr_var=None,
                param_fixed=None,
                refyvar=0,
            )
            out = prop.perform_checks(
                functiond,
                xsd,
                xerrsd,
                corr_x=corrd,
                repeat_dims=[0, 1],
                corr_dims=1,
                output_vars=2,
                fixed_corr_var=None,
                param_fixed=[True, False, False],
                refyvar=0,
            )
        except:
            print("done")


if __name__ == "__main__":
    unittest.main()
